import argparse
from pathlib import Path

import yaml


class ArgumentParser(argparse.ArgumentParser):
    """Simple implementation of ArgumentParser supporting config file

    This class is originated from https://github.com/bw2/ConfigArgParse,
    but this class is lack of some features that it has.

    - Not supporting multiple config files
    - Automatically adding "--config" as an option.
    - Not supporting any formats other than yaml
    - Not checking argument type

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_argument("--config", help="Give config file in yaml format")

    def parse_known_args(self, args=None, namespace=None):
        # Once parsing for setting from "--config"
        _args, _ = super().parse_known_args(args, namespace)
        if _args.config is not None:
            if not Path(_args.config).exists():
                self.error(f"No such file: {_args.config}")

            with open(_args.config, "r", encoding="utf-8") as f:
                d = yaml.safe_load(f)
            if not isinstance(d, dict):
                self.error("Config file has non dict value: {_args.config}")

            for key in d:
                for action in self._actions:
                    if key == action.dest:
                        break
                else:
                    self.error(f"unrecognized arguments: {key} (from {_args.config})")

            # NOTE(kamo): Ignore "--config" from a config file
            # NOTE(kamo): Unlike "configargparse", this module doesn't check type.
            #   i.e. We can set any type value regardless of argument type.
            self.set_defaults(**d)
        return super().parse_known_args(args, namespace)



from dataclasses import dataclass
import logging
import json
import os
import traceback
from typing import Any, Dict, List, Optional, Union

from achatbot.processors.llm.base import LLMProcessor
from achatbot.schemas.function_schema import FunctionSchema
from achatbot.schemas.tools_schema import ToolsSchema
from achatbot.common.event import EventHandlerManager
from achatbot.types.ai_conf import MCPServerConfig

try:
    from mcp import ClientSession, StdioServerParameters
    from mcp.client.session_group import SseServerParameters, StreamableHttpParameters
    from mcp.client.stdio import stdio_client
    from mcp.client.sse import sse_client
    from mcp.client.streamable_http import streamablehttp_client
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error("In order to use an MCP client, you need to `pip install achatbot[mcp]`.")
    raise Exception(f"Missing module: {e}")

# NOTE: for short session case, list_tools and call_tool with client session wrap, not long session


class MultiMCPClients(EventHandlerManager):
    def __init__(
        self,
        mcp_servers_config: Dict[str, MCPServerConfig],
    ):
        super().__init__()
        self.regist_clients: Dict[str, MCPClient] = {}
        for name, config in mcp_servers_config.items():
            if config.transport == "stdio":
                server_params = StdioServerParameters(**config.parameters)
                server_params.env = (
                    {**os.environ, **server_params.env} if server_params.env else {**os.environ}
                )
            elif config.transport == "sse":
                server_params = SseServerParameters(**config.parameters)
            elif config.transport == "streamable-http":
                server_params = StreamableHttpParameters(**config.parameters)
            else:
                logging.error(f"Unknown transport type: {config.transport}")
                continue
            mcp_client = MCPClient(server_params=server_params, mcp_name=name)
            self.regist_clients[name] = mcp_client
            logging.info(f"Registered mcp client: {name}")

    async def register_tools(self, llm: LLMProcessor) -> ToolsSchema:
        all_standard_tools = []
        for _, client in self.regist_clients.items():
            tools_schema = await client.register_tools(llm)
            tools_schema.standard_tools and all_standard_tools.extend(tools_schema.standard_tools)
            # todo: custom tools e.g.:{"adapter_type": [{"tool_name": "tool_description"}, ...], ...}"}
        return ToolsSchema(standard_tools=all_standard_tools)


class MCPClient(EventHandlerManager):
    def __init__(
        self,
        server_params: Union[StdioServerParameters, SseServerParameters, StreamableHttpParameters],
        **kwargs,
    ):
        super().__init__()
        self._mcp_name = kwargs.get("mcp_name", "mcp")
        self._server_params = server_params
        self._session = ClientSession
        if isinstance(server_params, StdioServerParameters):
            self._client = stdio_client
        elif isinstance(server_params, SseServerParameters):
            self._client = sse_client
            self._register_tools = self._sse_register_tools
        elif isinstance(server_params, StreamableHttpParameters):
            self._client = streamablehttp_client
            self._register_tools = self._streamable_http_register_tools
        else:
            raise TypeError(
                f"{self} invalid argument type: `server_params` must be either StdioServerParameters or an SSE server url string."
            )

    async def register_tools(self, llm: LLMProcessor) -> ToolsSchema:
        tools_schema = await self._register_tools(llm)
        return tools_schema

    def _convert_mcp_schema(self, tool_name: str, tool_schema: Dict[str, Any]) -> FunctionSchema:
        """Convert an mcp tool schema to FunctionSchema format.
        Args:
            tool_name: The name of the tool
            tool_schema: The mcp tool schema
        Returns:
            A FunctionSchema instance
        """

        logging.debug(f"Converting schema for tool '{tool_name}'")
        logging.trace(f"Original schema: {json.dumps(tool_schema, indent=2)}")

        properties = tool_schema["input_schema"].get("properties", {})
        required = tool_schema["input_schema"].get("required", [])

        schema = FunctionSchema(
            name=tool_name,
            description=tool_schema["description"],
            properties=properties,
            required=required,
        )

        logging.trace(f"Converted schema: {json.dumps(schema.to_default_dict(), indent=2)}")

        return schema

    async def _streamable_http_register_tools(self, llm: LLMProcessor) -> ToolsSchema:
        """Register all available streamable http mcp server tools with the LLM processor.
        Args:
            llm: The achatbot LLM processor to register tools with
        Returns:
            A ToolsSchema containing all registered tools
        """

        async def mcp_tool_wrapper(
            function_name: str,
            tool_call_id: str,
            arguments: Dict[str, Any],
            llm: LLMProcessor,
            context: any,
            result_callback: any,
        ) -> None:
            logging.debug(f"Executing tool '{function_name}' with call ID: {tool_call_id}")
            logging.trace(f"Tool arguments: {json.dumps(arguments, indent=2)}")
            try:
                async with self._client(
                    url=self._server_params.url,
                    headers=self._server_params.headers,
                    timeout=self._server_params.timeout,
                    sse_read_timeout=self._server_params.sse_read_timeout,
                    terminate_on_close=self._server_params.terminate_on_close,
                    auth=None,  # todo if use auth mcp server(or third party auth server)
                ) as (read, write, _):
                    async with self._session(read, write) as session:
                        await session.initialize()
                        await self._call_tool(session, function_name, arguments, result_callback)
            except Exception as e:
                error_msg = f"Error calling mcp {self._mcp_name} tool {function_name}: {str(e)}"
                logging.error(error_msg)
                logging.exception("Full exception details:")
                await result_callback(error_msg)

        async with self._client(
            url=self._server_params.url,
            headers=self._server_params.headers,
            timeout=self._server_params.timeout,
            sse_read_timeout=self._server_params.sse_read_timeout,
            terminate_on_close=self._server_params.terminate_on_close,
            auth=None,  # todo if use auth mcp server(or third party auth server)
        ) as (read, write, _):
            async with self._session(read, write) as session:
                await session.initialize()
                tools_schema = await self._list_tools(session, mcp_tool_wrapper, llm)
                return tools_schema

    async def _sse_register_tools(self, llm: LLMProcessor) -> ToolsSchema:
        """Register all available mcp sse server tools with the LLM processor.
        Args:
            llm: The achatbot LLM processor to register tools with
        Returns:
            A ToolsSchema containing all registered tools
        """

        async def mcp_tool_wrapper(
            function_name: str,
            tool_call_id: str,
            arguments: Dict[str, Any],
            llm: LLMProcessor,
            context: any,
            result_callback: any,
        ) -> None:
            logging.debug(f"Executing tool '{function_name}' with call ID: {tool_call_id}")
            logging.trace(f"Tool arguments: {json.dumps(arguments, indent=2)}")
            try:
                async with self._client(
                    url=self._server_params.url,
                    headers=self._server_params.headers,
                    timeout=self._server_params.timeout,
                    sse_read_timeout=self._server_params.sse_read_timeout,
                    auth=None,  # todo if use auth mcp server(or third party auth server)
                ) as (read, write):
                    async with self._session(read, write) as session:
                        await session.initialize()
                        await self._call_tool(session, function_name, arguments, result_callback)
            except Exception as e:
                error_msg = f"Error calling mcp {self._mcp_name} tool {function_name}: {str(e)}"
                logging.error(error_msg)
                logging.exception("Full exception details:")
                await result_callback(error_msg)

        async with self._client(
            url=self._server_params.url,
            headers=self._server_params.headers,
            timeout=self._server_params.timeout,
            sse_read_timeout=self._server_params.sse_read_timeout,
            auth=None,  # todo if use auth mcp server(or third party auth server)
        ) as (read, write):
            async with self._session(read, write) as session:
                await session.initialize()
                tools_schema = await self._list_tools(session, mcp_tool_wrapper, llm)
                return tools_schema

    async def _register_tools(self, llm: LLMProcessor) -> ToolsSchema:
        """Register all available mcp streamable http server tools with the LLM processor.
        Args:
            llm: The achatbot LLM processor to register tools with
        Returns:
            A ToolsSchema containing all registered tools
        """

        async def mcp_tool_wrapper(
            function_name: str,
            tool_call_id: str,
            arguments: Dict[str, Any],
            llm: LLMProcessor,
            context: any,
            result_callback: any,
        ) -> None:
            logging.debug(f"Executing tool '{function_name}' with call ID: {tool_call_id}")
            logging.trace(f"Tool arguments: {json.dumps(arguments, indent=2)}")
            try:
                async with self._client(self._server_params) as (read, write):
                    async with self._session(read, write) as session:
                        await session.initialize()
                        await self._call_tool(session, function_name, arguments, result_callback)
            except Exception as e:
                error_msg = f"Error calling mcp {self._mcp_name} tool {function_name}: {str(e)}"
                logging.error(error_msg)
                logging.exception("Full exception details:")
                await result_callback(error_msg)

        async with self._client(self._server_params) as (read, write):
            async with self._session(read, write) as session:
                await session.initialize()
                tools_schema = await self._list_tools(session, mcp_tool_wrapper, llm)
                return tools_schema

    async def _call_tool(self, session, function_name, arguments, result_callback):
        logging.debug(f"Calling mcp tool '{function_name}'")
        try:
            results = await session.call_tool(function_name, arguments=arguments)
        except Exception as e:
            error_msg = f"Error calling mcp tool {function_name}: {str(e)}"
            logging.error(error_msg)

        response = ""
        if results:
            if hasattr(results, "content") and results.content:
                for i, content in enumerate(results.content):
                    if hasattr(content, "text") and content.text:
                        logging.debug(f"Tool response chunk {i}: {content.text}")
                        response += content.text
                    else:
                        # logging.debug(f"Non-text result content: '{content}'")
                        pass
                logging.info(f"MCP {self._mcp_name} Tool '{function_name}' completed successfully")
                logging.debug(f"Final response: {response}")
            else:
                logging.error(
                    f"MCP {self._mcp_name}Error getting content from {function_name} results."
                )

        final_response = response if len(response) else "Sorry, could not call the mcp tool"
        await result_callback(final_response)

    async def _list_tools(self, session: ClientSession, mcp_tool_wrapper, llm: LLMProcessor):
        available_tools = await session.list_tools()
        if available_tools is None or available_tools.tools is None:
            logging.warning(f"MCP {self._mcp_name} No tools found")
            return ToolsSchema(standard_tools=[])

        tool_schemas: List[FunctionSchema] = []
        for tool in available_tools.tools:
            tool_name = tool.name
            try:
                # Convert the schema
                function_schema = self._convert_mcp_schema(
                    tool_name,
                    {"description": tool.description, "input_schema": tool.inputSchema},
                )

                if llm.has_function(tool_name):
                    logging.warning(
                        f"MCP {self._mcp_name} {tool_name} is already registered in LLMProcessor, over registering"
                    )
                # Register the wrapped function
                llm.register_function(tool_name, mcp_tool_wrapper)

                # Add to list of schemas
                tool_schemas.append(function_schema)
                logging.info(
                    f"MCP {self._mcp_name} Successfully registered tool '{tool_name}' | Tool description: {tool.description} | Tool schema: {tool.inputSchema}"
                )

            except Exception as e:
                logging.error(
                    f"MCP {self._mcp_name} Failed to register tool '{tool_name}': {str(e)}"
                )
                logging.exception("Full exception details:")
                continue

        logging.info(f"MCP {self._mcp_name} Completed registration of {len(tool_schemas)} tools")
        # {
        #    logging.info(f"tool_schemas[{i}]: {str(item)}") for i, item in enumerate(tool_schemas)
        # } if tool_schemas else logging.info(tool_schemas)
        tools_schema = ToolsSchema(standard_tools=tool_schemas)

        return tools_schema



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""isort:skip_file"""

import argparse
import importlib
import os

from fairseq.dataclass import FairseqDataclass
from fairseq.dataclass.utils import merge_with_parent, populate_dataclass
from hydra.core.config_store import ConfigStore

from .fairseq_task import FairseqTask, LegacyFairseqTask  # noqa


# register dataclass
TASK_DATACLASS_REGISTRY = {}
TASK_REGISTRY = {}
TASK_CLASS_NAMES = set()


def setup_task(cfg: FairseqDataclass, **kwargs):
    task = None
    task_name = getattr(cfg, "task", None)

    if isinstance(task_name, str):
        # legacy tasks
        task = TASK_REGISTRY[task_name]
        if task_name in TASK_DATACLASS_REGISTRY:
            dc = TASK_DATACLASS_REGISTRY[task_name]
            cfg = populate_dataclass(dc(), cfg)
    else:
        task_name = getattr(cfg, "_name", None)

        if task_name and task_name in TASK_DATACLASS_REGISTRY:
            dc = TASK_DATACLASS_REGISTRY[task_name]
            cfg = merge_with_parent(dc(), cfg)
            task = TASK_REGISTRY[task_name]

    assert (
        task is not None
    ), f"Could not infer task type from {cfg}. Available tasks: {TASK_REGISTRY.keys()}"

    return task.setup_task(cfg, **kwargs)


def register_task(name, dataclass=None):
    """
    New tasks can be added to fairseq with the
    :func:`~fairseq.tasks.register_task` function decorator.

    For example::

        @register_task('classification')
        class ClassificationTask(FairseqTask):
            (...)

    .. note::

        All Tasks must implement the :class:`~fairseq.tasks.FairseqTask`
        interface.

    Args:
        name (str): the name of the task
    """

    def register_task_cls(cls):
        if name in TASK_REGISTRY:
            raise ValueError("Cannot register duplicate task ({})".format(name))
        if not issubclass(cls, FairseqTask):
            raise ValueError(
                "Task ({}: {}) must extend FairseqTask".format(name, cls.__name__)
            )
        if cls.__name__ in TASK_CLASS_NAMES:
            raise ValueError(
                "Cannot register task with duplicate class name ({})".format(
                    cls.__name__
                )
            )
        TASK_REGISTRY[name] = cls
        TASK_CLASS_NAMES.add(cls.__name__)

        if dataclass is not None and not issubclass(dataclass, FairseqDataclass):
            raise ValueError(
                "Dataclass {} must extend FairseqDataclass".format(dataclass)
            )

        cls.__dataclass = dataclass
        if dataclass is not None:
            TASK_DATACLASS_REGISTRY[name] = dataclass

            cs = ConfigStore.instance()
            node = dataclass()
            node._name = name
            cs.store(name=name, group="task", node=node, provider="fairseq")

        return cls

    return register_task_cls


def get_task(name):
    return TASK_REGISTRY[name]


def import_tasks(tasks_dir, namespace):
    for file in os.listdir(tasks_dir):
        path = os.path.join(tasks_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            task_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + task_name)

            # expose `task_parser` for sphinx
            if task_name in TASK_REGISTRY:
                parser = argparse.ArgumentParser(add_help=False)
                group_task = parser.add_argument_group("Task name")
                # fmt: off
                group_task.add_argument('--task', metavar=task_name,
                                        help='Enable this task with: ``--task=' + task_name + '``')
                # fmt: on
                group_args = parser.add_argument_group(
                    "Additional command-line arguments"
                )
                TASK_REGISTRY[task_name].add_args(group_args)
                globals()[task_name + "_parser"] = parser


# automatically import any Python files in the tasks/ directory
tasks_dir = os.path.dirname(__file__)
import_tasks(tasks_dir, "fairseq.tasks")



import logging
import math
from typing import Optional, Tuple
from contextlib import nullcontext

from easydict import EasyDict as adict
import torch
from torch.nn import functional as F
from torch import nn

# use https://github.com/Dao-AILab/flash-attention
from flash_attn import flash_attn_qkvpacked_func, flash_attn_func

# use Optimus and Megatron Attention/Norm/Softmax op

# from optimus import flash_attn_func
# from megatron.model import LayerNorm
# from megatron.core import tensor_parallel
# from megatron.core import parallel_state as mpu
# from megatron.core.utils import make_viewless_tensor, divide
# from megatron.model.fused_rms_norm import RMSNorm
# from megatron.model.transformer import (
#     FlashSelfAttention,
#     NoopTransformerLayer,
#     _cfg_to_kwargs,
# )
# from megatron.model.enums import AttnMaskType, AttnType
# from megatron.model.fused_softmax import FusedScaleMaskSoftmax
# from megatron.model.utils import attention_mask_func

# from megatron.model.module import MegatronModule

# try:
#     from einops import rearrange
# except ImportError:
#     rearrange = None

# from flash_attn import flash_attn_varlen_func as flash_attn_unpadded_func

# try:
#     # flash attention 2.x
#     from flash_attn import flash_attn_varlen_func as flash_attn_unpadded_func
# except ImportError:
#     try:
#         # flash attention 1.x
#         from flash_attn.flash_attn_interface import flash_attn_unpadded_func
#     except ImportError:
#         flash_attn_unpadded_func = None

# try:
#     from flash_attn.flash_attn_interface import flash_attn_unpadded_relative_attention_bias_func
# except ImportError:
#     flash_attn_unpadded_relative_attention_bias_func = None

# try:
#     from flash_attn.flash_attn_interface import mask_flash_attn_unpadded_func
# except ImportError:
#     mask_flash_attn_unpadded_func = None


class LayerNormfp32(torch.nn.LayerNorm):
    """Subclass torch's LayerNorm to handle fp16."""

    def forward(self, x: torch.Tensor):
        orig_type = x.dtype
        ret = super().forward(x.type(torch.float32))
        return ret.type(orig_type)


def get_abs_pos(abs_pos, tgt_size):
    # abs_pos: L, C
    # tgt_size: M
    # return: M, C

    # print(tgt_size)
    # print(abs_pos.shape)
    # exit()
    dim = abs_pos.size(-1)
    # print(dim)
    abs_pos_new = abs_pos.squeeze(0)
    cls_token, old_pos_embed = abs_pos_new[:1], abs_pos_new[1:]

    src_size = int(math.sqrt(abs_pos_new.shape[0] - 1))
    tgt_size = int(math.sqrt(tgt_size))
    dtype = abs_pos.dtype

    if src_size != tgt_size:
        old_pos_embed = (
            old_pos_embed.view(1, src_size, src_size, dim).permute(0, 3, 1, 2).contiguous()
        )
        old_pos_embed = old_pos_embed.to(torch.float32)
        new_pos_embed = F.interpolate(
            old_pos_embed,
            size=(tgt_size, tgt_size),
            mode="bicubic",
            antialias=True,
            align_corners=False,
        ).to(dtype)
        new_pos_embed = new_pos_embed.permute(0, 2, 3, 1)
        new_pos_embed = new_pos_embed.view(tgt_size * tgt_size, dim)
        vision_pos_embed = torch.cat([cls_token, new_pos_embed], dim=0)
        vision_pos_embed = vision_pos_embed.view(1, tgt_size * tgt_size + 1, dim)
        return vision_pos_embed
    else:
        return abs_pos


@torch.jit.script
def quick_gelu(x):
    return x * torch.sigmoid(1.702 * x)


class CLIPVisionEmbeddings(nn.Module):
    def __init__(self, hidden_size=1024, image_size=224, patch_size=14, num_channels=3):
        super().__init__()
        self.embed_dim = hidden_size
        self.image_size = image_size
        self.patch_size = patch_size

        self.class_embedding = torch.nn.Parameter(torch.randn(self.embed_dim))

        self.patch_embedding = torch.nn.Conv2d(
            in_channels=num_channels,
            out_channels=self.embed_dim,
            kernel_size=self.patch_size,
            stride=self.patch_size,
            bias=False,
        )

        self.num_patches = (self.image_size // self.patch_size) ** 2
        self.num_positions = self.num_patches + 1
        self.position_embedding = torch.nn.Embedding(self.num_positions, self.embed_dim)
        self.register_buffer("position_ids", torch.arange(self.num_positions).expand((1, -1)))

    def forward(self, pixel_values, patch_embeds):
        batch_size = pixel_values.shape[0]
        # patch_embeds = self.patch_embedding(
        #     pixel_values
        # )  # shape = [*, width, grid, grid]

        if patch_embeds is not None:
            patch_embeds = patch_embeds
            # print(patch_embeds.shape)
        else:
            patch_embeds = self.patch_embedding(pixel_values)
            # print(111111)
        # shape = [*, width, grid, grid]
        # patch_embeds = patch_embeds.flatten(2).transpose(1, 2)

        patch_embeds = patch_embeds.flatten(2).transpose(1, 2)

        class_embeds = self.class_embedding.expand(batch_size, 1, -1)
        embeddings = torch.cat([class_embeds, patch_embeds], dim=1)

        # x = torch.cat([cls_token, x], dim=1)
        embeddings = embeddings + get_abs_pos(
            self.position_embedding(self.position_ids), embeddings.size(1)
        )
        # embeddings = embeddings + self.position_embedding(self.position_ids)
        return embeddings


class NoTPFeedForward(nn.Module):
    def __init__(
        self,
        cfg,
        dim: int,
        hidden_dim: int,
    ):
        super().__init__()

        self.fc1 = torch.nn.Linear(dim, hidden_dim, bias=True)
        self.fc2 = torch.nn.Linear(hidden_dim, dim, bias=True)

    def forward(self, x):
        output = self.fc2(quick_gelu(self.fc1(x)))
        return output


# from optimus.flash_attn_interface import flash_attn_qkvpacked_func


# class NoTPAttention(nn.Module):
#     def __init__(self, cfg):
#         super().__init__()
#         self.num_heads = cfg.num_attention_heads
#         self.n_local_heads = cfg.num_attention_heads
#         self.head_dim = cfg.hidden_size // cfg.num_attention_heads
#         self.max_seq_len = cfg.seq_length
#         self.use_flash_attention = cfg.use_flash_attn

#         self.qkv_proj = torch.nn.Linear(cfg.hidden_size, cfg.hidden_size * 3, bias=True)
#         self.out_proj = torch.nn.Linear(cfg.hidden_size, cfg.hidden_size, bias=True)

#         # self.core_attention = CoreAttention(cfg, AttnType.self_attn)

#         self.attn_drop = cfg.attention_dropout

#     def forward(
#             self,
#             x: torch.Tensor,
#     ):
#         bsz, seqlen, _ = x.shape
#         xqkv = self.qkv_proj(x)
#         xqkv = xqkv.view(bsz, seqlen, 3, self.num_heads, self.head_dim)

#         if self.use_flash_attention:
#             output = flash_attn_qkvpacked_func(xqkv)
#             output = output.view(bsz, seqlen, -1)
#         else:
#             xq, xk, xv = torch.split(xqkv, 1, dim=2)
#             xq = xq.squeeze(2)
#             xk = xk.squeeze(2)
#             xv = xv.squeeze(2)
#             # xq, xk, xv = xqkv[:, :, 0, ...], xqkv[:, :, 1, ...], xqkv[:, :, 2, ...]

#             # （B, num_head, S, head_size)
#             xq = xq.permute(0, 2, 1, 3)
#             xk = xk.permute(0, 2, 1, 3)
#             xv = xv.permute(0, 2, 1, 3)

#             output = torch.nn.functional.scaled_dot_product_attention(xq, xk, xv, attn_mask=None)
#             utput = output.permute(0, 2, 1, 3).view(bsz, seqlen, -1)
#         output = self.out_proj(output)
#         return output


# from optimus.flash_attn_interface import flash_attn_qkvpacked_func


class NoTPAttention(torch.nn.Module):
    def __init__(self, cfg):
        super().__init__()
        self.num_heads = cfg.num_attention_heads
        self.n_local_heads = cfg.num_attention_heads
        self.head_dim = cfg.hidden_size // cfg.num_attention_heads
        self.max_seq_len = cfg.seq_length
        self.use_flash_attention = cfg.use_flash_attn

        self.qkv_proj = torch.nn.Linear(cfg.hidden_size, cfg.hidden_size * 3, bias=True)
        self.out_proj = torch.nn.Linear(cfg.hidden_size, cfg.hidden_size, bias=True)

        # self.core_attention = CoreAttention(cfg, AttnType.self_attn)

        self.attn_drop = cfg.attention_dropout

    def forward(
        self,
        x: torch.Tensor,
    ):
        bsz, seqlen, _ = x.shape
        xqkv = self.qkv_proj(x)
        xqkv = xqkv.view(bsz, seqlen, 3, self.num_heads, self.head_dim)

        if self.use_flash_attention:
            output = flash_attn_qkvpacked_func(xqkv)
            output = output.view(bsz, seqlen, -1)
            # xq, xk, xv = torch.split(xqkv, 1, dim=2)
            # xq = xq.squeeze(2)
            # xk = xk.squeeze(2)
            # xv = xv.squeeze(2)
            # # xq, xk, xv = xqkv[:, :, 0, ...], xqkv[:, :, 1, ...], xqkv[:, :, 2, ...]

            # # （B, num_head, S, head_size)
            # xq = xq.permute(0, 2, 1, 3)
            # xk = xk.permute(0, 2, 1, 3)
            # xv = xv.permute(0, 2, 1, 3)
            # # with torch.backends.cuda.sdp_kernel(enable_flash=True, enable_math=False, enable_mem_efficient=False):
            # output = torch.nn.functional.scaled_dot_product_attention(xq, xk, xv, attn_mask=None)
            # output = output.permute(0, 2, 1, 3).reshape(bsz, seqlen, -1)
            # output = output.permute(0, 2, 1, 3).contiguous().view(bsz, seqlen, -1)
        else:
            # output = flash_attn_qkvpacked_func(xqkv)
            xq, xk, xv = torch.split(xqkv, 1, dim=2)
            xq = xq.squeeze(2)
            xk = xk.squeeze(2)
            xv = xv.squeeze(2)
            # xq, xk, xv = xqkv[:, :, 0, ...], xqkv[:, :, 1, ...], xqkv[:, :, 2, ...]

            # （B, num_head, S, head_size)
            xq = xq.permute(0, 2, 1, 3)
            xk = xk.permute(0, 2, 1, 3)
            xv = xv.permute(0, 2, 1, 3)
            # with torch.backends.cuda.sdp_kernel(enable_flash=True, enable_math=False, enable_mem_efficient=False):
            output = torch.nn.functional.scaled_dot_product_attention(xq, xk, xv, attn_mask=None)
            output = output.permute(0, 2, 1, 3).reshape(bsz, seqlen, -1)
        output = self.out_proj(output)
        return output


class NoTPTransformerBlock(nn.Module):
    def __init__(self, cfg, layer_id: int, multiple_of=256):
        super().__init__()

        self.n_heads = cfg.num_attention_heads
        self.dim = cfg.hidden_size
        self.head_dim = cfg.hidden_size // cfg.num_attention_heads
        self.self_attn = NoTPAttention(cfg)
        self.mlp = NoTPFeedForward(cfg, dim=cfg.hidden_size, hidden_dim=cfg.ffn_hidden_size)
        self.layer_id = layer_id
        self.layer_norm1 = torch.nn.LayerNorm(cfg.hidden_size, eps=cfg.layernorm_epsilon)
        self.layer_norm2 = torch.nn.LayerNorm(cfg.hidden_size, eps=cfg.layernorm_epsilon)

    def forward(self, x: torch.Tensor):
        residual = self.self_attn.forward(self.layer_norm1(x))
        h = x + residual
        out = h + self.mlp.forward(self.layer_norm2(h))
        return out


class NoTPTransformer(nn.Module):
    def __init__(self, cfg):
        super().__init__()

        self.cfg = cfg
        # self.recompute_list = self.cfg.get("recompute_list", [])
        self.num_layers = cfg.num_layers  # _get_num_layers(cfg)

        self.layers = torch.nn.ModuleList()
        for layer_id in range(self.num_layers):
            self.layers.append(
                NoTPTransformerBlock(
                    cfg,
                    layer_id + 1,
                )
            )

    def forward(
        self,
        hidden_states,
    ):
        for lid, layer in enumerate(self.layers):
            # if lid in self.recompute_list:
            #     def custom(layer_id):
            #         def custom_forward(*args, **kwargs):
            #             x_ = self.layers[layer_id](*args, **kwargs)
            #             return x_

            #         return custom_forward

            #     assert hidden_states.requires_grad == True, logger.warning(
            #         "When using recalculation, the input must have grad fn"
            #     )
            #     hidden_states = tensor_parallel.checkpoint(
            #         custom(lid),
            #         False,
            #         hidden_states.contiguous()
            #     )
            # else:
            hidden_states = layer(hidden_states)

        return hidden_states


# from megatron.core.tensor_parallel.layers import non_tensor_paralleled, local_dp_reduce, local_dp_scatter


class VitModel(nn.Module):
    def __init__(self, cfg, freeze_embed=False, freeze_pre_norm=False) -> None:
        super().__init__()

        self.embeddings = CLIPVisionEmbeddings(
            hidden_size=cfg.hidden_size, image_size=cfg.image_size, patch_size=cfg.patch_size
        )

        if freeze_embed:
            for name, param in self.embeddings.named_parameters():
                param.requires_grad = False

        self.transformer = NoTPTransformer(cfg=cfg)

        if cfg.get("fp32norm", False):
            logging.info("Load fp32 layernorm for ViT.")
            self.pre_layrnorm = LayerNormfp32(
                cfg.hidden_size,
                eps=cfg.get("pre_layernorm_epsilon", 1e-5),
            )
        else:
            self.pre_layrnorm = torch.nn.LayerNorm(
                cfg.hidden_size,
                eps=cfg.get("pre_layernorm_epsilon", 1e-5),
            )

        # self.pre_layrnorm = RMSNorm(
        #     cfg.hidden_size,
        #     eps=cfg.get("pre_layernorm_epsilon", 1e-5),
        #     sequence_parallel=False,
        #     use_fp32=True,
        #     use_optimus=True,
        # )

        if freeze_pre_norm:
            for name, param in self.pre_layrnorm.named_parameters():
                param.requires_grad = False

        for p in self.parameters():
            p.micro_dp = True

    def set_input_tensor(self, input_tensor):
        if not isinstance(input_tensor, list):
            input_tensor = [input_tensor]
        self.transformer.set_input_tensor(input_tensor[0])

    def __str__(self) -> str:
        return "open_clip"

    def forward(self, x, patch_embeds):
        x = self.embeddings(x, patch_embeds)
        hidden_states = self.pre_layrnorm(x)

        # hidden_states, dis = local_dp_scatter(hidden_states)
        output = self.transformer(hidden_states)

        # output = local_dp_reduce(output, dis)

        return output


vit_model_cfg = adict(
    num_layers=24,
    hidden_size=1024,
    num_heads=16,
    num_attention_heads=16,
    ffn_hidden_size=4096,
    seq_length=256,
    max_position_embeddings=256,
    use_flash_attn=False,
    understand_projector_stride=2,
    hidden_dropout=0.0,
    attention_dropout=0.0,
    no_persist_layer_norm=False,
    layernorm_epsilon=1e-5,
    pre_layernorm_epsilon=1e-5,
    image_size=224,
    patch_size=14,
    recompute_list=[],
)


def build_clip_l():
    return VitModel(
        cfg=vit_model_cfg,
        freeze_embed=False,
        freeze_pre_norm=False,
    )


if __name__ == "__main__":
    from mmgpt.model.vision_encoder.sam_b import build_sam_vit_b

    vit_model_cfg = adict(
        num_layers=24,
        hidden_size=1024,
        num_attention_heads=16,
        ffn_hidden_size=4096,
        seq_length=256,
        max_position_embeddings=256,
        use_flash_attn=False,
        understand_projector_stride=2,
        hidden_dropout=0.0,
        attention_dropout=0.0,
        no_persist_layer_norm=False,
        layernorm_epsilon=1e-5,
        pre_layernorm_epsilon=1e-5,
        image_size=224,
        patch_size=14,
        recompute_list=[],
    )

    sam_model = build_sam_vit_b()

    vision_model = VitModel(
        cfg=vit_model_cfg,
        freeze_embed=False,
        freeze_pre_norm=False,
    )

    # model = VitModel(1344)
    # x = torch.zeros(2, 3, 224, 224)
    x = torch.zeros(2, 3, 1024, 1024)

    with torch.no_grad():
        # y = vision_model(x)
        patch_embed = sam_model(x)
        print(patch_embed.shape)
        y = vision_model(x, patch_embed)
        print(y.shape)

        image_feature = torch.add(y[:, 1:], patch_embed.flatten(2).permute(0, 2, 1))

        print(image_feature.shape)



import os
import time
import traceback
import wave

import numpy as np
from pyrnnoise import RNNoise
import typer

from src.common.utils.audio_resample import (
    resample_bytes2bytes,
    resample_file,
    resample_bytes2numpy,
)

app = typer.Typer()


@app.command()
def random_np_test():
    denoiser = RNNoise(sample_rate=48000)

    try:
        i = 0
        for i in range(10000):
            audio_data = np.random.randint(-32768, 32767, (1, 512), dtype=np.int16)
            is_last = i == 1000
            for speech_prob, denoised_audio in denoiser.denoise_chunk(audio_data, is_last):
                print(f"{denoised_audio.shape}")
                print(f"Speech probability: {speech_prob}")
                # Process denoised_audio as needed
            if is_last is True:
                denoiser.reset()
    except Exception as e:
        print(i, e, traceback.format_exc())


@app.command()
def file_test_48k():
    denoiser = RNNoise(sample_rate=48000)
    in_path = "./records/speech_with_noise48k.wav"
    out_path = "./records/output_rnnoise_enhancer_stream_48k.wav"
    for speech_prob in denoiser.denoise_wav(in_path, out_path):
        print(speech_prob)
    print(f"{in_path=} save to {out_path}")


@app.command()
def file_test_16k():
    """
    NOTE:
    - 16K to 48K end save 16k denoise not good
    - need 16k to 48k resample first
    """
    denoiser = RNNoise(sample_rate=48000)
    raw_path = "./records/speech_with_noise16k.wav"
    in_path = "./records/speech_with_noise16k_48k.wav"
    resample_file(raw_path, in_path)
    out_path = "./records/output_rnnoise_enhancer_stream16k_48k.wav"
    for speech_prob in denoiser.denoise_wav(in_path, out_path):
        print(speech_prob)
    print(f"{in_path=} save to {out_path}")


@app.command()
def online_stream():
    denoiser = RNNoise(sample_rate=48000)

    # 模型每次处理的音频片段不能小于block_size
    # NOTE:
    # - rnnoise min support 10ms for 48000 sample rate
    # - if not 10ms frame size, AudioGraph to buffering until 10ms frame size
    # SEE:
    # - https://av.basswood-io.com/docs/stable/api/filter.html#bv.filter.graph.Graph
    # https://ffmpeg.org/ffmpeg-filters.html#abuffer

    # in_audio_path = "./records/speech_with_noise_48k.wav"
    # block_bytes_size = 480 * 2  # 10ms
    # block_bytes_size = 48 * 2  # 1ms

    in_audio_path = "./records/speech_with_noise16k.wav"
    # block_bytes_size = 160 * 2
    # block_bytes_size = 16 * 2
    block_bytes_size = 512 * 2  # silero vad frame size * 2

    start = time.time()
    waveform_48k_int16 = resample_bytes2numpy(
        b" " * block_bytes_size,
        orig_freq=16000,
        new_freq=48000,
    )
    denoiser.denoise_chunk(waveform_48k_int16, False)
    print(f"warmup cost: {time.time() - start} s")

    for i in range(10):
        denoiser.reset()
        audio_pcm = b""
        begin = time.time()
        with open(os.path.join(os.getcwd(), in_audio_path), "rb") as f:
            # 跳过 WAV 头
            f.read(44)

            audio_bytes = f.read(block_bytes_size)
            while len(audio_bytes) > 0:
                start = time.time()
                print("before", len(audio_bytes))
                waveform_48k_int16 = resample_bytes2numpy(
                    audio_bytes,
                    orig_freq=16000,
                    new_freq=48000,
                )
                pcm = b""
                for speech_prob, frame in denoiser.denoise_chunk(
                    waveform_48k_int16, len(audio_bytes) != block_bytes_size
                ):
                    # print(speech_prob)
                    pcm_48k = frame.tobytes()
                    pcm += pcm_48k
                pcm = resample_bytes2bytes(
                    pcm,
                    orig_freq=48000,
                    new_freq=16000,
                )
                print("after", len(pcm))
                audio_pcm += pcm
                print(f"cost: {time.time() - start} s")
                audio_bytes = f.read(block_bytes_size)

    print(f"total cost: {time.time() - begin} s")
    out_path = "./records/output_rnnoise_enhancer_stream.wav"
    with wave.open(out_path, "wb") as wf:
        wf.setnchannels(1)
        wf.setsampwidth(2)
        wf.setframerate(16000)
        wf.writeframes(audio_pcm)
        print(f"{in_audio_path=} save to {out_path}")


@app.command()
def online_stream_48k():
    denoiser = RNNoise(sample_rate=48000)

    # 模型每次处理的音频片段不能小于block_size
    # NOTE:
    # - rnnoise min support 10ms for 48000 sample rate
    # - if not 10ms frame size, AudioGraph to buffering until 10ms frame size
    # SEE:
    # - https://av.basswood-io.com/docs/stable/api/filter.html#bv.filter.graph.Graph
    # https://ffmpeg.org/ffmpeg-filters.html#abuffer

    in_audio_path = "./records/speech_with_noise_48k.wav"
    block_bytes_size = 480 * 2  # 10ms
    # block_bytes_size = 48 * 2  # 1ms

    start = time.time()
    denoiser.denoise_chunk(b" " * block_bytes_size, False)
    print(f"warmup cost: {time.time() - start} s")

    for i in range(10):
        denoiser.reset()
        try:
            audio_pcm = b""
            begin = time.time()
            with open(os.path.join(os.getcwd(), in_audio_path), "rb") as f:
                # 跳过 WAV 头
                f.read(44)

                audio_bytes = f.read(block_bytes_size)
                while len(audio_bytes) > 0:
                    start = time.time()
                    print("before", len(audio_bytes))
                    pcm = b""
                    # 1. 转换为 int16 数组
                    waveform_int16 = np.frombuffer(audio_bytes, dtype=np.int16)
                    # 2. 转换为归一化浮点数
                    # waveform_float = waveform_int16.astype(np.float32) / 32768.0
                    for speech_prob, frame in denoiser.denoise_chunk(
                        waveform_int16, len(audio_bytes) != block_bytes_size
                    ):
                        # print(speech_prob)
                        pcm_48k = frame.tobytes()
                        pcm += pcm_48k
                    print("after", len(pcm))
                    audio_pcm += pcm
                    print(f"cost: {time.time() - start} s")
                    audio_bytes = f.read(block_bytes_size)
        except Exception as e:
            print(i, e, traceback.format_exc())
            break

    print(f"total cost: {time.time() - begin} s")

    if len(audio_pcm) > 0:
        out_path = "./records/output_rnnoise_enhancer_stream_48k.wav"
        with wave.open(out_path, "wb") as wf:
            wf.setnchannels(1)
            wf.setsampwidth(2)
            wf.setframerate(48000)
            wf.writeframes(audio_pcm)
            print(f"{in_audio_path=} save to {out_path}")


"""
python -m demo.ans_rnnoise_enhancer random-np-test
python -m demo.ans_rnnoise_enhancer file-test-16k
python -m demo.ans_rnnoise_enhancer file-test-48k
python -m demo.ans_rnnoise_enhancer online-stream
python -m demo.ans_rnnoise_enhancer online-stream-48k
"""
if __name__ == "__main__":
    app()



from enum import Enum
from typing import Any, Optional
from pydantic import BaseModel


class AvatarStatus(Enum):
    IDEL = 0
    LISTENING = 1
    THINKING = 2
    RESPONDING = 3
    SPEAKING = 4

    def __str__(self) -> str:
        if self == AvatarStatus.IDEL:
            return "Idel"
        if self == AvatarStatus.LISTENING:
            return "Listening"
        if self == AvatarStatus.THINKING:
            return "Thinking"
        if self == AvatarStatus.RESPONDING:
            return "Responding"
        if self == AvatarStatus.SPEAKING:
            return "Speaking"
        return self.name


class AudioSlice(BaseModel):
    speech_id: Any
    play_audio_data: bytes
    play_audio_sample_rate: int
    algo_audio_data: Optional[bytes]
    algo_audio_sample_rate: int
    end_of_speech: bool
    front_padding_duration: float = 0
    end_padding_duration: float = 0

    def get_audio_duration(self) -> float:
        return len(self.play_audio_data) / self.play_audio_sample_rate / 2

    def get_algo_audio_duration(self) -> float:
        return (
            len(self.algo_audio_data) / self.algo_audio_sample_rate / 2
            if self.algo_audio_data
            else 0
        )

    def __str__(self):
        return f"{self.speech_id=} {self.get_audio_duration()=} {self.get_algo_audio_duration()=} {self.end_of_speech=} {self.front_padding_duration=} {self.end_padding_duration=}"


class SpeechAudio(BaseModel):
    """
    only support mono audio for now
    """

    end_of_speech: bool = False
    speech_id: Any = ""
    sample_rate: int = 16000
    audio_data: bytes = bytes()

    def get_audio_duration(self):
        return len(self.audio_data) / self.sample_rate / 2



import asyncio
import logging
from typing import Literal

from src.common.task_manager.base import TaskManager
from src.common.task_manager.multiprocessing_task_manager import MultiprocessingTaskManager
from src.common.task_manager.asyncio_task_manager import AsyncioTaskManager
from src.common.task_manager.threading_task_manager import ThreadingTaskManager


class TaskManagerFactory:
    loop: asyncio.AbstractEventLoop | None = None

    @staticmethod
    def task_manager(
        type: Literal["multiprocessing", "threading", "asyncio"] = "multiprocessing",
        task_done_timeout=5,
    ) -> TaskManager:
        match type:
            case "asyncio":
                return AsyncioTaskManager(
                    task_done_timeout=task_done_timeout, loop=TaskManagerFactory.loop
                )
            case "threading":
                return ThreadingTaskManager(task_done_timeout=task_done_timeout)
            case "multiprocessing":
                return MultiprocessingTaskManager(task_done_timeout=task_done_timeout)
            case _:
                return MultiprocessingTaskManager(task_done_timeout=task_done_timeout)

    @staticmethod
    def cleanup(func):
        try:
            if asyncio.iscoroutinefunction(func):
                if TaskManagerFactory.loop is None or TaskManagerFactory.loop.is_closed():
                    TaskManagerFactory.loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(TaskManagerFactory.loop)

                TaskManagerFactory.loop.run_until_complete(func())
            else:
                func()

            if TaskManagerFactory.loop:
                TaskManagerFactory.loop.close()
        except Exception as e:
            logging.error(f"Error during TaskManagerFactory cleanup: {e}")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .distributed_timeout_wrapper import DistributedTimeoutWrapper
from .fully_sharded_data_parallel import fsdp_enable_wrap, fsdp_wrap, FullyShardedDataParallel
from .legacy_distributed_data_parallel import LegacyDistributedDataParallel
from .module_proxy_wrapper import ModuleProxyWrapper


__all__ = [
    "DistributedTimeoutWrapper",
    "fsdp_enable_wrap",
    "fsdp_wrap",
    "FullyShardedDataParallel",
    "LegacyDistributedDataParallel",
    "ModuleProxyWrapper",
]



import os
import asyncio
import logging
from typing import List, Optional
from concurrent.futures import ThreadPoolExecutor

from apipeline.processors.async_frame_processor import FrameDirection
from apipeline.frames import CancelFrame, StartFrame, EndFrame, Frame, TextFrame, ErrorFrame
from a2a.utils.message import new_agent_text_message
from google.adk.events.event import Event as ADKEvent
from google.adk.artifacts import BaseArtifactService
from google.adk.memory.base_memory_service import BaseMemoryService
from google.adk.sessions.base_session_service import BaseSessionService

from src.processors.session_processor import SessionProcessor
from src.services.a2a_multiagents.adk_host_agent_manager import ADKHostAgentManager
from src.services.help.httpx import HTTPXClientWrapper
from src.common.session import Session
from src.types.frames import LLMMessagesFrame, LLMFullResponseStartFrame, LLMFullResponseEndFrame
from src.processors.aggregators.openai_llm_context import (
    OpenAILLMContext,
    OpenAILLMContextFrame,
)


class A2AConversationProcessor(SessionProcessor):
    def __init__(
        self,
        app_name: str,
        api_key: str = "",
        mode: str = "supervisor",  # supervisor or planer
        system_prompt: str = "",
        http_timeout: float = 30.0,
        agent_urls: Optional[list[str]] = [],  # agent http url
        max_workers: int = 0,
        session: Session | None = None,
        session_service: BaseSessionService | None = None,
        memory_service: BaseMemoryService | None = None,
        artifact_service: BaseArtifactService | None = None,
        **kwargs,
    ):
        super().__init__(session=session, **kwargs)
        self.http_client_wrapper = HTTPXClientWrapper()
        self.http_client_wrapper.start(timeout=http_timeout)
        http_client = self.http_client_wrapper()
        self.manager = ADKHostAgentManager(
            http_client,
            app_name,
            api_key=api_key,
            mode=mode,
            remote_agent_addresses=agent_urls,
            system_prompt=system_prompt,
            session_service=session_service,
            memory_service=memory_service,
            artifact_service=artifact_service,
        )
        self.manager.initialize_host()
        self.queue = asyncio.Queue()
        self.push_task: Optional[asyncio.Task] = None
        self.running = False
        self.conversation = None

        max_workers = max_workers or os.cpu_count()
        self.executor = ThreadPoolExecutor(max_workers=max_workers)

    @property
    def host_agent_manager(self):
        return self.manager

    def set_user_id(self, user_id: str):
        self.manager.set_user_id(user_id)

    async def create_conversation(self):
        self.conversation = await self.manager.create_conversation()

    async def register_agents(self, urls: List[str]):
        for url in urls:
            await self.manager.register_agent(url)

    async def start(self, frame: StartFrame):
        await self.create_conversation()
        self.running = True
        self.push_task = self.get_event_loop().create_task(self._push_frames())

        if self.http_client_wrapper is None:
            self.http_client_wrapper = HTTPXClientWrapper()
            self.http_client_wrapper.start()
            self.manager.set_client(self.http_client_wrapper())

        logging.info(f"{self.name} Conversation started")

    async def stop(self, frame: EndFrame):
        self.running = False
        if self.http_client_wrapper:
            await self.http_client_wrapper.stop()
            self.http_client_wrapper = None

        logging.info(f"{self.name} Conversation end")

    async def cancel(self, frame: CancelFrame):
        self.push_task.cancel()
        try:
            await self.push_task
        except asyncio.CancelledError:
            logging.info(f"{self.name} push_task cancelled.")
        if self.http_client_wrapper:
            await self.http_client_wrapper.stop()
            self.http_client_wrapper = None
        self.executor.shutdown(wait=True)
        logging.info(f"{self.name} Conversation cancelled")

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        """Process incoming frames, intercept context frames for memory integration.

        Args:
            frame: The incoming frame to process.
            direction: The direction of frame flow in the pipeline.
        """
        await super().process_frame(frame, direction)

        if isinstance(frame, StartFrame):
            await self.start(frame)
            await self.push_frame(frame, direction)
        elif isinstance(frame, EndFrame):
            await self.stop(frame)
            await self.push_frame(frame, direction)
        elif isinstance(frame, CancelFrame):
            await self.cancel(frame)
            await self.push_frame(frame, direction)
        else:
            await self.queue_frame(frame, direction)

        context = None
        messages = None
        if isinstance(frame, LLMMessagesFrame):
            messages = frame.messages
            context = OpenAILLMContext(messages=messages)
        elif isinstance(frame, OpenAILLMContextFrame):
            context = frame.context
        if not context:
            return
        # Get the latest user message to use as a query for a2a agents
        context_messages = context.get_messages()
        for message in reversed(context_messages):
            if message.get("role") == "user" and isinstance(message.get("content"), str):
                query: str = message.get("content")
                await self.send_message(query)
                break

    async def send_message(self, text: str):
        try:
            message = new_agent_text_message(
                text=text,
                context_id=self.conversation.conversation_id,
                # task_id=str(uuid.uuid4()),# new message no task_id
            )
            message = self.manager.sanitize_message(message)
            # print("message-->", message)
            # Send the message in a separate thread to avoid blocking the event loop.
            self.executor.submit(
                self.manager.process_message_threadsafe, message, self.get_event_loop(), self.queue
            )
            # t = threading.Thread(
            #    target=lambda: self.manager.process_message_threadsafe(
            #        message, self.get_event_loop(), self.queue
            #    )
            # )
            # t.start()
        except Exception as e:
            logging.error(f"Error processing: {str(e)}")
            await self.push_frame(ErrorFrame(f"Error processing: {str(e)}"))

    async def _push_frames(self):
        while self.running:
            try:
                event: ADKEvent = await asyncio.wait_for(self.queue.get(), timeout=1.0)
                # print("event-->", event)
                # Send the message in a separate thread to avoid blocking the event loop.
                if event.id and event.partial is None:
                    await self.queue_frame(LLMFullResponseStartFrame())
                if event.content and event.content.parts and event.partial:
                    for part in event.content.parts:
                        if part.text:
                            await self.queue_frame(
                                TextFrame(text=part.text), FrameDirection.DOWNSTREAM
                            )
                if event.content and event.content.parts and event.partial is False:
                    await self.queue_frame(LLMFullResponseEndFrame())

                self.queue.task_done()
            except asyncio.TimeoutError:
                continue
            except asyncio.CancelledError:
                logging.info(f"{self.name} _push_frames cancelled")
                break
            except Exception as ex:
                logging.exception(f"{self.name} Unexpected error in _push_frames: {ex}")
                if self.get_event_loop().is_closed():
                    logging.warning(f"{self.name} event loop is closed")
                    break



# Grouped Temporal Convolutional Recurrent Network (GTCRN)
import logging
import os
import time

import numpy as np
import onnxruntime
from librosa import istft, stft

from achatbot.common.utils.audio_resample import resample_bytes2bytes, resample_numpy2bytes
from achatbot.common.utils.audio_utils import bytes2NpArrayWith16
from achatbot.common.session import Session, SessionCtx
from achatbot.common.types import MODELS_DIR
from achatbot.common.factory import EngineClass
from achatbot.common.interface import ISpeechEnhancer


class GTCRNSpeechEnhancer(EngineClass, ISpeechEnhancer):
    """
    onnx realtime
    - https://github.com/Xiaobin-Rong/gtcrn
    - wget https://github.com/Xiaobin-Rong/gtcrn/raw/refs/heads/main/stream/onnx_models/gtcrn_simple.onnx -O ./models/gtcrn_simple.onnx
    """

    TAG = "enhancer_ans_gtcrn_onnx"
    SAMPLE_RATE = 16000

    def __init__(self, **kwargs):
        super().__init__()
        model = kwargs.get("model", os.path.join(MODELS_DIR, "gtcrn_simple.onnx"))

        self.reset()
        self.warmup_cn = kwargs.get("warmup_cn", 1)
        self.infer_session = onnxruntime.InferenceSession(
            model, None, providers=["CPUExecutionProvider"]
        )

    def warmup(self, session: Session, **kwargs):
        if self.warmup_cn <= 0:
            return
        sample_rate = kwargs.get("sample_rate") or session.ctx.sampling_rate
        assert sample_rate, "sample_rate is None or 0"

        start = time.time()
        x = bytes2NpArrayWith16(
            b" " * (sample_rate * 2 // 100),  # 10ms
        )
        x = stft(
            x,
            n_fft=512,
            hop_length=256,
            win_length=512,
            window=np.hanning(512) ** 0.5,
            center=True,
            pad_mode="constant",
            dtype=None,
        )

        inputs = np.stack([x.real, x.imag], axis=0)[None, ...].astype("float32")
        inputs = np.transpose(inputs, (0, 2, 3, 1))  # (1,F,Time,2)

        conv_cache = np.zeros([2, 1, 16, 16, 33], dtype="float32")
        tra_cache = np.zeros([2, 3, 1, 1, 16], dtype="float32")
        inter_cache = np.zeros([2, 1, 33, 16], dtype="float32")
        outputs = []
        for i in range(inputs.shape[-2]):
            out_i, conv_cache, tra_cache, inter_cache = self.infer_session.run(
                [],
                {
                    "mix": inputs[..., i : i + 1, :],
                    "conv_cache": conv_cache,
                    "tra_cache": tra_cache,
                    "inter_cache": inter_cache,
                },
            )
            outputs.append(out_i)

        outputs = np.concatenate(outputs, axis=2)
        _ = istft(
            outputs[..., 0] + 1j * outputs[..., 1],
            n_fft=512,
            hop_length=256,
            win_length=512,
            window=np.hanning(512) ** 0.5,
        )

        logging.info(f"{self.TAG} warmup cost: {time.time() - start} s")

    def enhance(self, session: Session, **kwargs):
        sample_rate = kwargs.get("sample_rate") or session.ctx.sampling_rate
        assert sample_rate, "sample_rate is None or 0"
        audio_chunk = session.ctx.state.get("audio_chunk", b"")
        is_last = session.ctx.state.get("is_last", False)

        if len(audio_chunk) == 0:
            return audio_chunk
        if sample_rate != self.SAMPLE_RATE:
            audio_chunk = resample_bytes2bytes(
                audio_chunk,
                orig_freq=sample_rate,
                new_freq=self.SAMPLE_RATE,
            )

        x = bytes2NpArrayWith16(audio_chunk)
        x = stft(
            x,
            n_fft=512,
            hop_length=256,
            win_length=512,
            window=np.hanning(512) ** 0.5,
            center=True,
            pad_mode="constant",
            dtype=None,
        )

        inputs = np.stack([x.real, x.imag], axis=0)[None, ...].astype("float32")
        inputs = np.transpose(inputs, (0, 2, 3, 1))  # (1,F,Time,2)

        outputs = []
        for i in range(inputs.shape[-2]):
            out_i, self.conv_cache, self.tra_cache, self.inter_cache = self.infer_session.run(
                [],
                {
                    "mix": inputs[..., i : i + 1, :],
                    "conv_cache": self.conv_cache,
                    "tra_cache": self.tra_cache,
                    "inter_cache": self.inter_cache,
                },
            )
            outputs.append(out_i)

        outputs = np.concatenate(outputs, axis=2)
        out_pcm = istft(
            outputs[..., 0] + 1j * outputs[..., 1],
            n_fft=512,
            hop_length=256,
            win_length=512,
            window=np.hanning(512) ** 0.5,
        )
        out_pcm_bytes = b""
        if len(out_pcm) > 0:
            out_pcm_bytes = resample_numpy2bytes(
                out_pcm,
                orig_freq=self.SAMPLE_RATE,
                new_freq=sample_rate,
            )
        if is_last is True:
            self.reset()

        return out_pcm_bytes

    def reset(self):
        self.conv_cache = np.zeros([2, 1, 16, 16, 33], dtype="float32")
        self.tra_cache = np.zeros([2, 3, 1, 1, 16], dtype="float32")
        self.inter_cache = np.zeros([2, 1, 33, 16], dtype="float32")



from typing import Awaitable, Callable

from apipeline.frames.sys_frames import Frame
from apipeline.processors.user_idle_processor import UserIdleProcessor
from apipeline.processors.frame_processor import FrameDirection

from achatbot.types.frames.control_frames import (
    BotSpeakingFrame,
    UserStartedSpeakingFrame,
    UserStoppedSpeakingFrame,
)


class UserIdleProcessor(UserIdleProcessor):
    """This class is based UserIdleProcessor,
    for BotSpeakingFrame,UserStartedSpeakingFrame,UserStoppedSpeakingFrame

    timeout to call the callback.
    """

    def __init__(
        self,
        *,
        callback: Callable[["UserIdleProcessor"], Awaitable[None]],
        timeout: float,
        **kwargs,
    ):
        super().__init__(callback=callback, timeout=timeout, **kwargs)

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        # We shouldn't call the idle callback if the user or the bot are speaking.
        if isinstance(frame, UserStartedSpeakingFrame):
            self._interrupted = True
            self._idle_event.set()
        elif isinstance(frame, UserStoppedSpeakingFrame):
            self._interrupted = False
            self._idle_event.set()
        elif isinstance(frame, BotSpeakingFrame):
            self._idle_event.set()



from dataclasses import dataclass
import os

from achatbot.types.speech.asr.base import ASRArgs
from achatbot.common.types import ASSETS_DIR, MODELS_DIR


@dataclass
class WhisperTensorRTLLMASRArgs(ASRArgs):
    use_py_session: bool = True
    log_level: str = "info"

    engine_dir: str = os.path.join(MODELS_DIR, "whisper", "trt_engines_float16")
    assets_dir: str = ASSETS_DIR
    max_input_len: int = 3000
    max_output_len: int = 96
    max_batch_size: int = 1  # need <= build decoder_model_config.max_batch_size
    kv_cache_free_gpu_memory_fraction: float = 0.9
    cross_kv_cache_fraction: float = 0.5
    num_beams: int = 1
    debug: bool = False

    text_prefix: str = "<|startoftranscript|><|en|><|transcribe|><|notimestamps|>"
    """
    1. max: pad to the 30s, using the option if the model is trained with max padding e.g. openai official models,
    2. longest: pad to the longest sequence in the batch,
    3. nopad: no padding, only works with cpp session,
    """
    padding_strategy: str = "max"



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from torch import nn


class SamePad(nn.Module):
    def __init__(self, kernel_size, causal=False):
        super().__init__()
        if causal:
            self.remove = kernel_size - 1
        else:
            self.remove = 1 if kernel_size % 2 == 0 else 0

    def forward(self, x):
        if self.remove > 0:
            x = x[:, :, : -self.remove]
        return x



"""
The code is base on https://github.com/Pointcept/Pointcept
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from .builder import LOSSES


@LOSSES.register_module()
class CrossEntropyLoss(nn.Module):
    def __init__(
        self,
        weight=None,
        size_average=None,
        reduce=None,
        reduction="mean",
        label_smoothing=0.0,
        loss_weight=1.0,
        ignore_index=-1,
    ):
        super(CrossEntropyLoss, self).__init__()
        weight = torch.tensor(weight).cuda() if weight is not None else None
        self.loss_weight = loss_weight
        self.loss = nn.CrossEntropyLoss(
            weight=weight,
            size_average=size_average,
            ignore_index=ignore_index,
            reduce=reduce,
            reduction=reduction,
            label_smoothing=label_smoothing,
        )

    def forward(self, pred, target):
        return self.loss(pred, target) * self.loss_weight


@LOSSES.register_module()
class L1Loss(nn.Module):
    def __init__(
        self,
        weight=None,
        size_average=None,
        reduce=None,
        reduction="mean",
        label_smoothing=0.0,
        loss_weight=1.0,
        ignore_index=-1,
    ):
        super(L1Loss, self).__init__()
        weight = torch.tensor(weight).cuda() if weight is not None else None
        self.loss_weight = loss_weight
        self.loss = nn.L1Loss(reduction='mean')

    def forward(self, pred, target):
        return self.loss(pred, target[:,None]) * self.loss_weight


@LOSSES.register_module()
class SmoothCELoss(nn.Module):
    def __init__(self, smoothing_ratio=0.1):
        super(SmoothCELoss, self).__init__()
        self.smoothing_ratio = smoothing_ratio

    def forward(self, pred, target):
        eps = self.smoothing_ratio
        n_class = pred.size(1)
        one_hot = torch.zeros_like(pred).scatter(1, target.view(-1, 1), 1)
        one_hot = one_hot * (1 - eps) + (1 - one_hot) * eps / (n_class - 1)
        log_prb = F.log_softmax(pred, dim=1)
        loss = -(one_hot * log_prb).total(dim=1)
        loss = loss[torch.isfinite(loss)].mean()
        return loss


@LOSSES.register_module()
class BinaryFocalLoss(nn.Module):
    def __init__(self, gamma=2.0, alpha=0.5, logits=True, reduce=True, loss_weight=1.0):
        """Binary Focal Loss
        <https://arxiv.org/abs/1708.02002>`
        """
        super(BinaryFocalLoss, self).__init__()
        assert 0 < alpha < 1
        self.gamma = gamma
        self.alpha = alpha
        self.logits = logits
        self.reduce = reduce
        self.loss_weight = loss_weight

    def forward(self, pred, target, **kwargs):
        """Forward function.
        Args:
            pred (torch.Tensor): The prediction with shape (N)
            target (torch.Tensor): The ground truth. If containing class
                indices, shape (N) where each value is 0≤targets[i]≤1, If containing class probabilities,
                same shape as the input.
        Returns:
            torch.Tensor: The calculated loss
        """
        if self.logits:
            bce = F.binary_cross_entropy_with_logits(pred, target, reduction="none")
        else:
            bce = F.binary_cross_entropy(pred, target, reduction="none")
        pt = torch.exp(-bce)
        alpha = self.alpha * target + (1 - self.alpha) * (1 - target)
        focal_loss = alpha * (1 - pt) ** self.gamma * bce

        if self.reduce:
            focal_loss = torch.mean(focal_loss)
        return focal_loss * self.loss_weight


@LOSSES.register_module()
class FocalLoss(nn.Module):
    def __init__(
        self, gamma=2.0, alpha=0.5, reduction="mean", loss_weight=1.0, ignore_index=-1
    ):
        """Focal Loss
        <https://arxiv.org/abs/1708.02002>`
        """
        super(FocalLoss, self).__init__()
        assert reduction in (
            "mean",
            "sum",
        ), "AssertionError: reduction should be 'mean' or 'sum'"
        assert isinstance(
            alpha, (float, list)
        ), "AssertionError: alpha should be of type float"
        assert isinstance(gamma, float), "AssertionError: gamma should be of type float"
        assert isinstance(
            loss_weight, float
        ), "AssertionError: loss_weight should be of type float"
        assert isinstance(ignore_index, int), "ignore_index must be of type int"
        self.gamma = gamma
        self.alpha = alpha
        self.reduction = reduction
        self.loss_weight = loss_weight
        self.ignore_index = ignore_index

    def forward(self, pred, target, **kwargs):
        """Forward function.
        Args:
            pred (torch.Tensor): The prediction with shape (N, C) where C = number of classes.
            target (torch.Tensor): The ground truth. If containing class
                indices, shape (N) where each value is 0≤targets[i]≤C−1, If containing class probabilities,
                same shape as the input.
        Returns:
            torch.Tensor: The calculated loss
        """
        # [B, C, d_1, d_2, ..., d_k] -> [C, B, d_1, d_2, ..., d_k]
        pred = pred.transpose(0, 1)
        # [C, B, d_1, d_2, ..., d_k] -> [C, N]
        pred = pred.reshape(pred.size(0), -1)
        # [C, N] -> [N, C]
        pred = pred.transpose(0, 1).contiguous()
        # (B, d_1, d_2, ..., d_k) --> (B * d_1 * d_2 * ... * d_k,)
        target = target.view(-1).contiguous()
        assert pred.size(0) == target.size(
            0
        ), "The shape of pred doesn't match the shape of target"
        valid_mask = target != self.ignore_index
        target = target[valid_mask]
        pred = pred[valid_mask]

        if len(target) == 0:
            return 0.0

        num_classes = pred.size(1)
        target = F.one_hot(target, num_classes=num_classes)

        alpha = self.alpha
        if isinstance(alpha, list):
            alpha = pred.new_tensor(alpha)
        pred_sigmoid = pred.sigmoid()
        target = target.type_as(pred)
        one_minus_pt = (1 - pred_sigmoid) * target + pred_sigmoid * (1 - target)
        focal_weight = (alpha * target + (1 - alpha) * (1 - target)) * one_minus_pt.pow(
            self.gamma
        )

        loss = (
            F.binary_cross_entropy_with_logits(pred, target, reduction="none")
            * focal_weight
        )
        if self.reduction == "mean":
            loss = loss.mean()
        elif self.reduction == "sum":
            loss = loss.total()
        return self.loss_weight * loss


@LOSSES.register_module()
class DiceLoss(nn.Module):
    def __init__(self, smooth=1, exponent=2, loss_weight=1.0, ignore_index=-1):
        """DiceLoss.
        This loss is proposed in `V-Net: Fully Convolutional Neural Networks for
        Volumetric Medical Image Segmentation <https://arxiv.org/abs/1606.04797>`_.
        """
        super(DiceLoss, self).__init__()
        self.smooth = smooth
        self.exponent = exponent
        self.loss_weight = loss_weight
        self.ignore_index = ignore_index

    def forward(self, pred, target, **kwargs):
        # [B, C, d_1, d_2, ..., d_k] -> [C, B, d_1, d_2, ..., d_k]
        pred = pred.transpose(0, 1)
        # [C, B, d_1, d_2, ..., d_k] -> [C, N]
        pred = pred.reshape(pred.size(0), -1)
        # [C, N] -> [N, C]
        pred = pred.transpose(0, 1).contiguous()
        # (B, d_1, d_2, ..., d_k) --> (B * d_1 * d_2 * ... * d_k,)
        target = target.view(-1).contiguous()
        assert pred.size(0) == target.size(
            0
        ), "The shape of pred doesn't match the shape of target"
        valid_mask = target != self.ignore_index
        target = target[valid_mask]
        pred = pred[valid_mask]

        pred = F.softmax(pred, dim=1)
        num_classes = pred.shape[1]
        target = F.one_hot(
            torch.clamp(target.long(), 0, num_classes - 1), num_classes=num_classes
        )

        total_loss = 0
        for i in range(num_classes):
            if i != self.ignore_index:
                num = torch.sum(torch.mul(pred[:, i], target[:, i])) * 2 + self.smooth
                den = (
                    torch.sum(
                        pred[:, i].pow(self.exponent) + target[:, i].pow(self.exponent)
                    )
                    + self.smooth
                )
                dice_loss = 1 - num / den
                total_loss += dice_loss
        loss = total_loss / num_classes
        return self.loss_weight * loss



import logging
import numpy as np
import torch
import torch.nn.functional as F
from torch import nn
from subprocess import CalledProcessError, run, Popen, PIPE
import os
from functools import lru_cache
from typing import Optional, Union
from .modeling_whisper import WhisperModel

# hard-coded audio hyperparameters
SAMPLE_RATE = 16000
N_FFT = 400
N_MELS = 120
HOP_LENGTH = 160
CHUNK_LENGTH = 30
N_SAMPLES = CHUNK_LENGTH * SAMPLE_RATE  # 480000 samples in a 30-second chunk


def load_bytesio_audio(content, sr: int = SAMPLE_RATE):
    cmd = [
        "ffmpeg",
        "-nostdin",
        "-threads",
        "0",
        "-i",
        "pipe:",
        "-f",
        "s16le",
        "-ac",
        "1",
        "-acodec",
        "pcm_s16le",
        "-ar",
        str(sr),
        "pipe:",
    ]
    p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, bufsize=-1)
    out, _ = p.communicate(input=content)
    return np.frombuffer(out, np.int16).flatten().astype(np.float32) / 32768.0


def load_audio(file: str, sr: int = SAMPLE_RATE):
    """
    Open an audio file and read as mono waveform, resampling as necessary

    Parameters
    ----------
    file: str
        The audio file to open

    sr: int
        The sample rate to resample the audio if necessary

    Returns
    -------
    A NumPy array containing the audio waveform, in float32 dtype.
    """

    # This launches a subprocess to decode audio while down-mixing
    # and resampling as necessary.  Requires the ffmpeg CLI in PATH.
    # fmt: off
    cmd = ["ffmpeg", "-nostdin", "-threads", "0", "-i", file, "-f", "s16le", "-ac", "1", "-acodec", "pcm_s16le", "-ar", str(sr), "-"]
    # fmt: on
    try:
        out = run(cmd, capture_output=True, check=True).stdout
    except CalledProcessError as e:
        raise RuntimeError(f"Failed to load audio: {e.stderr.decode()}") from e

    return np.frombuffer(out, np.int16).flatten().astype(np.float32) / 32768.0


def pad_or_trim(array, length: int = N_SAMPLES, *, axis: int = -1):
    """
    Pad or trim the audio array to N_SAMPLES, as expected by the encoder.
    """
    if torch.is_tensor(array):
        if array.shape[axis] > length:
            array = array.index_select(
                dim=axis, index=torch.arange(length, device=array.device)
            )

        if array.shape[axis] < length:
            pad_widths = [(0, 0)] * array.ndim
            pad_widths[axis] = (0, length - array.shape[axis])
            array = F.pad(array, [pad for sizes in pad_widths[::-1] for pad in sizes])
    else:
        if array.shape[axis] > length:
            array = array.take(indices=range(length), axis=axis)

        if array.shape[axis] < length:
            pad_widths = [(0, 0)] * array.ndim
            pad_widths[axis] = (0, length - array.shape[axis])
            array = np.pad(array, pad_widths)

    return array


@lru_cache(maxsize=None)
def mel_filters(device, n_mels: int = 128) -> torch.Tensor:
    """
    load the mel filterbank matrix for projecting STFT into a Mel spectrogram.
    Allows decoupling librosa dependency; saved using:

        np.savez_compressed(
            "mel_filters.npz",
            mel_80=librosa.filters.mel(sr=16000, n_fft=400, n_mels=80),
        )
    """
    default_path = os.path.join(os.path.dirname(__file__), "mel_filters.npz")
    mel_filters_path = os.getenv("MEL_FILTERS_PATH", default_path)
    with np.load( mel_filters_path) as f:
        logging.info(f"Loading mel filters from {mel_filters_path}")
        return torch.from_numpy(f[f"mel_{n_mels}"]).to(device)


def log_mel_spectrogram(
    audio: Union[str, np.ndarray, torch.Tensor],
    n_mels: int = 128,
    padding: int = 0,
    device: Optional[Union[str, torch.device]] = None,
):
    """
    Compute the log-Mel spectrogram of

    Parameters
    ----------
    audio: Union[str, np.ndarray, torch.Tensor], shape = (*)
        The path to audio or either a NumPy array or Tensor containing the audio waveform in 16 kHz

    n_mels: int
        The number of Mel-frequency filters, only 80 is supported

    padding: int
        Number of zero samples to pad to the right

    device: Optional[Union[str, torch.device]]
        If given, the audio tensor is moved to this device before STFT

    Returns
    -------
    torch.Tensor, shape = (80, n_frames)
        A Tensor that contains the Mel spectrogram
    """
    if not torch.is_tensor(audio):
        if isinstance(audio, str):
            audio = load_audio(audio)
        audio = torch.from_numpy(audio)

    if device is not None:
        audio = audio.to(device)
    if padding > 0:
        audio = F.pad(audio, (0, padding))
    window = torch.hann_window(N_FFT).to(audio.device)
    stft = torch.stft(audio, N_FFT, HOP_LENGTH, window=window, return_complex=True)
    magnitudes = stft[..., :-1].abs() ** 2

    filters = mel_filters(audio.device, n_mels)
    mel_spec = filters @ magnitudes

    log_spec = torch.clamp(mel_spec, min=1e-10).log10()
    log_spec = torch.maximum(log_spec, log_spec.max() - 8.0)
    log_spec = (log_spec + 4.0) / 4.0
    return log_spec


class WhisperEncoder(nn.Module):
    def __init__(
        self, model_path, mel_batch_size=40, unfreeze_online_whisper_model=False
    ):
        super().__init__()
        self.speech_encoder = WhisperModel.from_pretrained(model_path).encoder
        self.unfreeze_online_whisper_model = unfreeze_online_whisper_model
        if not self.unfreeze_online_whisper_model:
            self.speech_encoder.eval()
        self.mel_batch_size = mel_batch_size
        # self.speech_encoder.to(torch.cuda.current_device())
        # self.speech_encoder.half()

    @torch.no_grad()
    def tokenize_waveform(self, audio, kimia_whisper_clip_silence=False):
        if isinstance(audio, torch.Tensor):
            audio = audio[0]
            audio = audio.cpu().numpy()
        time_step = 0
        audios = []
        while time_step * 16000 < audio.shape[0]:
            audio_segment = audio[time_step * 16000 : (time_step + 30) * 16000]
            audios.append(audio_segment)
            time_step += 30

        final_audio_embedding = []

        for audio_segment in audios:
            # import pdb; pdb.set_trace()
            assert audio_segment.shape[0] <= 480000
            L = audio_segment.shape[0]
            token_len = (L - 1) // (
                160 * 8
            ) + 1  # to match huggingface logic, with use attention mask to control the length and the slice with mask[:, ::160], also match the glm4 12.5 logic

            pad_audio = pad_or_trim(audio_segment.flatten())
            mel = log_mel_spectrogram(pad_audio)  # torch.Size([80, 3000])
            assert mel.shape[1] == 3000
            if kimia_whisper_clip_silence:
                input_seq_lens_list = [token_len * 4]
                input_seq_lens = torch.LongTensor(input_seq_lens_list).to(
                    torch.cuda.current_device()
                )
                audio_embedding = self.speech_encoder(
                    mel.unsqueeze(0).to(torch.cuda.current_device()).to(torch.bfloat16),
                    return_dict=True,
                    input_seq_lens=input_seq_lens,
                ).last_hidden_state
            else:
                audio_embedding = self.speech_encoder(
                    mel.unsqueeze(0).to(torch.cuda.current_device()).to(torch.bfloat16),
                    return_dict=True,
                ).last_hidden_state
                # audio_embedding: [1, 3000, 1280]
                audio_embedding = audio_embedding[:, : token_len * 4, :].cpu()
            final_audio_embedding.append(audio_embedding)

        final_audio_embedding = torch.cat(final_audio_embedding, dim=1)
        return final_audio_embedding



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import tts_pb2 as tts__pb2


class TTSStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.LoadModel = channel.unary_unary(
                '/chat_bot.tts.TTS/LoadModel',
                request_serializer=tts__pb2.LoadModelRequest.SerializeToString,
                response_deserializer=tts__pb2.LoadModelResponse.FromString,
                )
        self.GetVoices = channel.unary_unary(
                '/chat_bot.tts.TTS/GetVoices',
                request_serializer=tts__pb2.GetVoicesRequest.SerializeToString,
                response_deserializer=tts__pb2.GetVoicesResponse.FromString,
                )
        self.SetVoice = channel.unary_unary(
                '/chat_bot.tts.TTS/SetVoice',
                request_serializer=tts__pb2.SetVoiceRequest.SerializeToString,
                response_deserializer=tts__pb2.SetVoiceResponse.FromString,
                )
        self.GetStreamInfo = channel.unary_unary(
                '/chat_bot.tts.TTS/GetStreamInfo',
                request_serializer=tts__pb2.GetStreamInfoRequest.SerializeToString,
                response_deserializer=tts__pb2.GetStreamInfoReponse.FromString,
                )
        self.SynthesizeUS = channel.unary_stream(
                '/chat_bot.tts.TTS/SynthesizeUS',
                request_serializer=tts__pb2.SynthesizeRequest.SerializeToString,
                response_deserializer=tts__pb2.SynthesizeResponse.FromString,
                )


class TTSServicer(object):
    """Missing associated documentation comment in .proto file."""

    def LoadModel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetVoices(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetVoice(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStreamInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SynthesizeUS(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TTSServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'LoadModel': grpc.unary_unary_rpc_method_handler(
                    servicer.LoadModel,
                    request_deserializer=tts__pb2.LoadModelRequest.FromString,
                    response_serializer=tts__pb2.LoadModelResponse.SerializeToString,
            ),
            'GetVoices': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVoices,
                    request_deserializer=tts__pb2.GetVoicesRequest.FromString,
                    response_serializer=tts__pb2.GetVoicesResponse.SerializeToString,
            ),
            'SetVoice': grpc.unary_unary_rpc_method_handler(
                    servicer.SetVoice,
                    request_deserializer=tts__pb2.SetVoiceRequest.FromString,
                    response_serializer=tts__pb2.SetVoiceResponse.SerializeToString,
            ),
            'GetStreamInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStreamInfo,
                    request_deserializer=tts__pb2.GetStreamInfoRequest.FromString,
                    response_serializer=tts__pb2.GetStreamInfoReponse.SerializeToString,
            ),
            'SynthesizeUS': grpc.unary_stream_rpc_method_handler(
                    servicer.SynthesizeUS,
                    request_deserializer=tts__pb2.SynthesizeRequest.FromString,
                    response_serializer=tts__pb2.SynthesizeResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'chat_bot.tts.TTS', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TTS(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def LoadModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/chat_bot.tts.TTS/LoadModel',
            tts__pb2.LoadModelRequest.SerializeToString,
            tts__pb2.LoadModelResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetVoices(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/chat_bot.tts.TTS/GetVoices',
            tts__pb2.GetVoicesRequest.SerializeToString,
            tts__pb2.GetVoicesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetVoice(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/chat_bot.tts.TTS/SetVoice',
            tts__pb2.SetVoiceRequest.SerializeToString,
            tts__pb2.SetVoiceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetStreamInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/chat_bot.tts.TTS/GetStreamInfo',
            tts__pb2.GetStreamInfoRequest.SerializeToString,
            tts__pb2.GetStreamInfoReponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SynthesizeUS(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/chat_bot.tts.TTS/SynthesizeUS',
            tts__pb2.SynthesizeRequest.SerializeToString,
            tts__pb2.SynthesizeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import os
import logging
import asyncio

import unittest

from src.common.utils.wav import save_audio_to_file
from src.common.factory import EngineFactory, EngineClass
from src.common.logger import Logger
from src.common.session import Session
from src.common.types import SessionCtx, MODELS_DIR, RECORDS_DIR
from src.modules.speech.tts.coqui_tts import CoquiTTS

r"""
REFERENCE_AUDIO_PATH= python -m unittest test.modules.speech.tts.test_coqui.TestCoquiTTS.test_synthesize
python -m unittest test.modules.speech.tts.test_coqui.TestCoquiTTS.test_synthesize_speak
STREAM=1 python -m unittest test.modules.speech.tts.test_coqui.TestCoquiTTS.test_synthesize_speak
"""


class TestCoquiTTS(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tts_tag = os.getenv("TTS_TAG", "tts_coqui")
        cls.tts_text = os.getenv("TTS_TEXT", "你好！")
        cls.stream = os.getenv("STREAM", "")
        cls.conf_file = os.getenv(
            "CONF_FILE", os.path.join(MODELS_DIR, "coqui/XTTS-v2/config.json")
        )
        cls.model_path = os.getenv("MODEL_PATH", os.path.join(MODELS_DIR, "coqui/XTTS-v2"))
        cls.reference_audio_path = os.getenv(
            "REFERENCE_AUDIO_PATH", os.path.join(RECORDS_DIR, "tmp.wav")
        )
        Logger.init(os.getenv("LOG_LEVEL", "debug").upper(), is_file=False)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        kwargs = {}
        kwargs["model_path"] = self.model_path
        kwargs["conf_file"] = self.conf_file
        kwargs["reference_audio_path"] = self.reference_audio_path
        self.tts: CoquiTTS = EngineFactory.get_engine_by_tag(EngineClass, self.tts_tag, **kwargs)
        self.session = Session(**SessionCtx("test_tts_client_id").__dict__)
        self.pyaudio_instance = None
        self.audio_stream = None

    def tearDown(self):
        self.audio_stream and self.audio_stream.stop_stream()
        self.audio_stream and self.audio_stream.close()
        self.pyaudio_instance and self.pyaudio_instance.terminate()

    def test_synthesize(self):
        stream_info = self.tts.get_stream_info()
        self.session.ctx.state["tts_text"] = self.tts_text
        print(self.session.ctx)
        iter = self.tts.synthesize_sync(self.session)
        res = bytearray()
        for i, chunk in enumerate(iter):
            print(i, len(chunk))
            res.extend(chunk)
        path = asyncio.run(
            save_audio_to_file(
                res,
                f"test_{self.tts.TAG}.wav",
                sample_rate=stream_info["rate"],
                sample_width=stream_info["sample_width"],
                channles=stream_info["channels"],
            )
        )
        print(path)

    def test_synthesize_speak(self):
        import pyaudio

        stream_info = self.tts.get_stream_info()
        self.pyaudio_instance = pyaudio.PyAudio()
        self.audio_stream = self.pyaudio_instance.open(
            format=stream_info["format"],
            channels=stream_info["channels"],
            rate=stream_info["rate"],
            output_device_index=None,
            output=True,
        )

        self.session.ctx.state["tts_text"] = self.tts_text
        print(self.session.ctx)
        self.tts.args.tts_stream = bool(self.stream)
        iter = self.tts.synthesize_sync(self.session)
        sub_chunk_size = 1024
        for i, chunk in enumerate(iter):
            print(f"get {i} chunk {len(chunk)}")
            if len(chunk) / sub_chunk_size < 100:
                self.audio_stream.write(chunk)
                continue
            for i in range(0, len(chunk), sub_chunk_size):
                sub_chunk = chunk[i : i + sub_chunk_size]
                self.audio_stream.write(sub_chunk)

    def test_get_voices(self):
        voices = self.tts.get_voices()
        self.assertGreater(len(voices), 0)
        print(voices, len(voices))



import torch.nn as nn

from models.losses import build_criteria
from .builder import MODELS, build_model

@MODELS.register_module()
class DefaultEstimator(nn.Module):
    def __init__(self, backbone=None, criteria=None):
        super().__init__()
        self.backbone = build_model(backbone)
        self.criteria = build_criteria(criteria)

    def forward(self, input_dict):
        pred_exp = self.backbone(input_dict)
        # train
        if self.training:
            loss = self.criteria(pred_exp, input_dict["gt_exp"])
            return dict(loss=loss)
        # eval
        elif "gt_exp" in input_dict.keys():
            loss = self.criteria(pred_exp, input_dict["gt_exp"])
            return dict(loss=loss, pred_exp=pred_exp)
        # infer
        else:
            return dict(pred_exp=pred_exp)



#!/usr/bin/python

# https://github.com/triton-inference-server/tensorrtllm_backend/blob/v0.18.0/tools/whisper/client.py

import asyncio
from collections import defaultdict
import logging
import os
import queue
import sys
import time
import types
from typing import Dict, Iterable, List, TextIO, Tuple
import uuid
import modal


app_name = "whisper"
app = modal.App(f"{app_name}-grpc-client")

# Define the dependencies for the function using a Modal Image.
image = modal.Image.debian_slim(python_version="3.10").apt_install("git", "wget")
image = image.pip_install(
    "numpy",
    "soundfile",
    "tritonclient[all]",
    "kaldialign",  # WER
)
image = image.run_commands(
    # god as a direct consequence of the sin which man thus punished had given her a lovely child whose place was on that same dishonoured bosom to connect her parent for ever with the race and descent of mortals and to be finally a blessed soul in heaven
    "wget 'https://raw.githubusercontent.com/yuekaizhang/Triton-ASR-Client/main/datasets/mini_en/wav/1221-135766-0001.wav' -O /1221-135766-0001.wav",
    # yet these thoughts affected hester prynne less with hope than apprehension
    "wget 'https://raw.githubusercontent.com/yuekaizhang/Triton-ASR-Client/main/datasets/mini_en/wav/1221-135766-0002.wav' -O /1221-135766-0002.wav",
    # 大学生利用漏洞免费吃肯德基获刑
    "wget 'https://raw.githubusercontent.com/yuekaizhang/Triton-ASR-Client/main/datasets/mini_zh/wav/long.wav' -O /long.wav",
    # 富士康在印度工厂出现大规模感染目前工厂产量已下降超50%
    "wget 'https://raw.githubusercontent.com/yuekaizhang/Triton-ASR-Client/main/datasets/mini_zh/wav/mid.wav' -O /mid.wav",
    "ls -lh /*.wav",
)
image = image.env({"TRITON_PROTOCOL": os.getenv("TRITON_PROTOCOL", "grpc")})

audio_text_map = {
    "/1221-135766-0001.wav": "god as a direct consequence of the sin which man thus punished had given her a lovely child whose place was on that same dishonoured bosom to connect her parent for ever with the race and descent of mortals and to be finally a blessed soul in heaven.",
    "/1221-135766-0002.wav": "yet these thoughts affected hester prynne less with hope than apprehension.",
    "/mid.wav": "大学生利用漏洞免费吃肯德基获刑。",
    "/long.wav": "富士康在印度工厂出现大规模感染，目前工厂产量已下降超50%。",
}

with image.imports():
    import numpy as np
    import soundfile as sf

    # import tritonclient.grpc as protocol_client
    protocol = os.getenv("TRITON_PROTOCOL", "grpc")
    print(f"use protocol: {protocol}")
    if protocol == "grpc":
        import tritonclient.grpc.aio as protocol_client
    else:
        import tritonclient.http.aio as protocol_client
    from tritonclient.grpc import InferenceServerException
    from tritonclient.utils import np_to_triton_dtype
    import kaldialign


ASSETS_DIR = "/root/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)


@app.function(
    cpu=2.0,
    retries=0,
    image=image,
    timeout=1200,  # default 300s
    scaledown_window=1200,
)
async def health(
    server_url: str,
    verbose: bool = False,
    model_names: str = "whisper_bls,whisper_tensorrt_llm",
):
    url = server_url

    try:
        triton_client = protocol_client.InferenceServerClient(url=url, verbose=verbose)
    except Exception as e:
        print("context creation failed: " + str(e))
        return

    # Health
    if not await triton_client.is_server_live(headers={"test": "1", "dummy": "2"}):
        print("FAILED : is_server_live")
        return
    if not await triton_client.is_server_ready():
        print("FAILED : is_server_ready")
        return

    # Metadata
    metadata = await triton_client.get_server_metadata()
    if isinstance(metadata, dict):
        metadata = types.SimpleNamespace(**metadata)
    if not (metadata.name == "triton"):
        print("FAILED : get_server_metadata")
        return
    print(metadata)

    # Health
    for model_name in model_names.split(","):
        if not await triton_client.is_model_ready(model_name):
            print(f"{model_name} FAILED : is_model_ready")

        print("-" * 20)

        metadata = await triton_client.get_model_metadata(
            model_name, headers={"test": "1", "dummy": "2"}
        )
        if isinstance(metadata, dict):
            metadata = types.SimpleNamespace(**metadata)
        if not (metadata.name == model_name):
            print(f"{model_name} FAILED : get_model_metadata")
            return
        print(metadata)


@app.function(
    cpu=2.0,
    retries=0,
    image=image,
    volumes={
        ASSETS_DIR: assets_dir,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
)
async def asr(
    server_url: str = "localhost:8000",
    reference_audio: str = "/1221-135766-0002.wav",
    text_prefix: str = "<|startoftranscript|><|en|><|transcribe|><|notimestamps|>",
    model_name: str = "whisper_bls",
    padding_duration: int = 10,
    streaming: bool = False,
    verbose: bool = False,
):
    triton_client = protocol_client.InferenceServerClient(url=server_url, verbose=verbose)
    waveform, sr = sf.read(reference_audio)
    duration = sf.info(reference_audio).duration
    assert sr == 16000, "sample rate hardcoded in server"

    samples = np.array(waveform, dtype=np.float32)
    inputs = prepare_grpc_sdk_request(
        samples, text_prefix, duration, sample_rate=sr, padding_duration=padding_duration
    )
    outputs = [protocol_client.InferRequestedOutput("TRANSCRIPTS")]

    if streaming:
        decoding_results = await infer_streaming(
            model_name,
            inputs,
            request_id=str(uuid.uuid4()),
            outputs=outputs,
            triton_client=triton_client,
        )
    else:
        decoding_results = await infer(
            model_name,
            inputs,
            request_id=str(uuid.uuid4()),
            outputs=outputs,
            triton_client=triton_client,
        )
    print(f"ref_text____________: {audio_text_map[reference_audio]}")
    print(f"asr_decoding_results: {decoding_results}")

    tot_err_rate = write_error_stats(
        None,
        test_set_name="test-asr",
        results=[
            (
                reference_audio,
                audio_text_map[reference_audio].lower(),
                decoding_results.strip().lower(),
            )
        ],
        enable_log=False,
    )
    print("total_err_rate:", tot_err_rate)

    await triton_client.close()


@app.function(
    cpu=8.0,
    retries=0,
    image=image,
    volumes={
        ASSETS_DIR: assets_dir,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
)
async def bench_asr(
    server_url: str = "localhost:8000",
    reference_audio: str = "/1221-135766-0002.wav",
    text_prefix: str = "<|startoftranscript|><|en|><|transcribe|><|notimestamps|>",
    model_name: str = "whisper_bls",
    padding_duration: int = 10,  # padding to nearset 10 seconds
    concurency_cn: int = 1,
    batch_size: int = 1,
    verbose: bool = False,
    streaming: bool = False,
):
    dps_list = [
        [
            {
                "wav_id": reference_audio,
                "ref_audio_filepath": reference_audio,
                "ref_text": audio_text_map[reference_audio],
            }
        ]
        * batch_size
    ] * concurency_cn

    triton_client = protocol_client.InferenceServerClient(url=server_url, verbose=verbose)

    tasks = []
    start_time = time.time()
    for i in range(concurency_cn):
        task = asyncio.create_task(
            send_whisper(
                dps=dps_list[i],
                name=f"task-{i}",
                triton_client=triton_client,
                protocol_client=protocol_client,
                log_interval=1,
                model_name=model_name,
                padding_duration=padding_duration,
                text_prefix=text_prefix,
                streaming=streaming,
            )
        )
        tasks.append(task)

    answer_list = await asyncio.gather(*tasks)

    end_time = time.time()
    elapsed = end_time - start_time

    results = []
    total_duration = 0.0
    latency_data = []
    for answer in answer_list:
        total_duration += answer[0]
        results += answer[1]
        latency_data += answer[2]

    rtf = elapsed / total_duration

    s = f"RTF: {rtf:.4f}\n"
    s += f"total_duration: {total_duration:.3f} seconds\n"
    s += f"({total_duration/3600:.2f} hours)\n"
    s += f"processing time: {elapsed:.3f} seconds " f"({elapsed/3600:.2f} hours)\n"

    latency_list = [chunk_end for (chunk_end, chunk_duration) in latency_data]
    latency_ms = sum(latency_list) / float(len(latency_list)) * 1000.0
    latency_variance = np.var(latency_list, dtype=np.float64) * 1000.0
    s += f"latency_variance: {latency_variance:.2f}\n"
    s += f"latency_50_percentile_ms: {np.percentile(latency_list, 50) * 1000.0:.2f}\n"
    s += f"latency_90_percentile_ms: {np.percentile(latency_list, 90) * 1000.0:.2f}\n"
    s += f"latency_95_percentile_ms: {np.percentile(latency_list, 95) * 1000.0:.2f}\n"
    s += f"latency_99_percentile_ms: {np.percentile(latency_list, 99) * 1000.0:.2f}\n"
    s += f"average_latency_ms: {latency_ms:.2f}\n"

    print(s)


def prepare_grpc_sdk_request(
    waveform,
    text_prefix,
    duration: float,
    sample_rate=16000,
    padding_duration: int = None,
):
    assert len(waveform.shape) == 1, "waveform should be 1D"
    lengths = np.array([[len(waveform)]], dtype=np.int32)
    if padding_duration:
        samples = np.zeros(
            (
                1,
                padding_duration * sample_rate * ((int(duration) // padding_duration) + 1),
            ),
            dtype=np.float32,
        )

        samples[0, : len(waveform)] = waveform
    else:
        samples = waveform

    samples = samples.reshape(1, -1).astype(np.float32)

    """
    data = {
        "inputs": [
            {
                "name": "TEXT_PREFIX",
                "shape": [1, 1],
                "datatype": "BYTES",
                "data": [TEXT_PREFIX],
            },
            {
                "name": "WAV",
                "shape": samples.shape,
                "datatype": "FP32",
                "data": samples.tolist(),
            },
            {
                "name": "WAV_LEN",
                "shape": lengths.shape,
                "datatype": "INT32",
                "data": lengths.tolist(),
            }
        ]
    }
    """
    inputs = [
        protocol_client.InferInput("TEXT_PREFIX", [1, 1], "BYTES"),
        protocol_client.InferInput("WAV", samples.shape, np_to_triton_dtype(samples.dtype)),
        protocol_client.InferInput("WAV_LEN", lengths.shape, np_to_triton_dtype(lengths.dtype)),
    ]
    input_data_numpy = np.array([text_prefix], dtype=object)
    input_data_numpy = input_data_numpy.reshape((1, 1))
    inputs[0].set_data_from_numpy(input_data_numpy)
    inputs[1].set_data_from_numpy(samples)
    inputs[2].set_data_from_numpy(lengths)

    return inputs


async def infer(
    model_name,
    inputs,
    request_id,
    outputs,
    triton_client: "protocol_client.InferenceServerClient",
):
    response = await triton_client.infer(model_name, inputs, request_id=request_id, outputs=outputs)
    decoding_results = response.as_numpy("TRANSCRIPTS")[0]
    if isinstance(decoding_results, np.ndarray):
        decoding_results = b" ".join(decoding_results).decode("utf-8")
    else:
        decoding_results = decoding_results.decode("utf-8")
    return decoding_results


class UserData:
    def __init__(self):
        self._completed_requests = queue.Queue()


async def infer_streaming(
    model_name,
    inputs,
    request_id,
    outputs,
    triton_client: "protocol_client.InferenceServerClient",
):
    user_data = UserData()

    async def async_request_iterator():
        yield {
            "model_name": model_name,
            "inputs": inputs,
            "outputs": outputs,
            "request_id": request_id,
        }

    try:
        response_iterator = triton_client.stream_infer(
            inputs_iterator=async_request_iterator(),
            stream_timeout=None,
        )
        async for response in response_iterator:
            result, error = response
            if error:
                print(error)
                user_data._completed_requests.put(error)
            else:
                user_data._completed_requests.put(result)
    except InferenceServerException as error:
        print(error)
        sys.exit(1)

    results = []
    while True:
        try:
            data_item = user_data._completed_requests.get(block=False)
            if isinstance(data_item, InferenceServerException):
                sys.exit(1)
            else:
                decoding_results = data_item.as_numpy("TRANSCRIPTS")[0]
                if isinstance(decoding_results, np.ndarray):
                    decoding_results = b" ".join(decoding_results).decode("utf-8")
                else:
                    decoding_results = decoding_results.decode("utf-8")
                results.append(decoding_results)
        except Exception:
            break
    decoding_results = results[-1]
    return decoding_results


async def send_whisper(
    dps: list,
    name: str,
    triton_client: "protocol_client.InferenceServerClient",
    protocol_client: types.ModuleType,
    log_interval: int,
    model_name: str,
    padding_duration: int = 10,  # padding to nearset 10 seconds
    text_prefix: str = "<|startoftranscript|><|en|><|transcribe|><|notimestamps|>",
    streaming: bool = False,
):
    total_duration = 0.0
    results = []
    latency_data = []
    task_id = int(name[5:])
    for i, dp in enumerate(dps):
        if i % log_interval == 0:
            print(f"{name}: {i}/{len(dps)}")

        waveform, sr = sf.read(dp["ref_audio_filepath"])
        assert sr == 16000, "sample rate hardcoded in server"
        duration = int(len(waveform) / sr)
        # duration = sf.info(dp["ref_audio_filepath"]).duration

        inputs = prepare_grpc_sdk_request(
            waveform, text_prefix, duration, sample_rate=sr, padding_duration=padding_duration
        )
        outputs = [protocol_client.InferRequestedOutput("TRANSCRIPTS")]

        # Send request
        sequence_id = 100000000 + i + task_id * 100
        start = time.time()
        if streaming:
            decoding_results = await infer_streaming(
                model_name,
                inputs,
                request_id=str(sequence_id),
                outputs=outputs,
                triton_client=triton_client,
            )
        else:
            decoding_results = await infer(
                model_name,
                inputs,
                request_id=str(sequence_id),
                outputs=outputs,
                triton_client=triton_client,
            )
        end = time.time() - start
        latency_data.append((end, duration))
        total_duration += duration
        results.append(
            (
                dp["wav_id"],
                dp["ref_text"].split(),  # label
                decoding_results.split(),  # prediction
            )
        )
        print(results[-1])

    return total_duration, results, latency_data


"""
# gRPC
## health check for whisper
modal run src/llm/trtllm/whisper/client.py \
    --action health \
    --model-name whisper \
    --model-names whisper \
    --server-url "r15.modal.host:34101"

## health check for whisper_bls,whisper_tensorrt_llm
modal run src/llm/trtllm/whisper/client.py \
    --action health \
    --server-url "r15.modal.host:34101"

## health check for whisper_infer_bls, whisper_tensorrt_llm_cpprunner
modal run src/llm/trtllm/whisper/client.py \
    --action health \
    --model-name whisper_infer_bls \
    --model-names whisper_infer_bls,whisper_tensorrt_llm_cpprunner \
    --server-url "r28.modal.host:38535"


## single wav test for whisper
modal run src/llm/trtllm/whisper/client.py \
    --no-streaming \
    --action asr \
    --model-name whisper \
    --server-url "r15.modal.host:34101"
modal run src/llm/trtllm/whisper/client.py \
    --streaming \
    --action asr \
    --model-name whisper \
    --server-url "r15.modal.host:34101"
modal run src/llm/trtllm/whisper/client.py \
    --action asr \
    --model-name whisper \
    --reference-audio /1221-135766-0001.wav \
    --text-prefix "<|startoftranscript|><|en|><|transcribe|><|notimestamps|>" \
    --server-url "r15.modal.host:34101"
modal run src/llm/trtllm/whisper/client.py \
    --action asr \
    --model-name whisper \
    --reference-audio /long.wav \
    --text-prefix "<|startoftranscript|><|zh|><|transcribe|><|notimestamps|>" \
    --server-url "r15.modal.host:34101"

## single wav test for whisper_bls,whisper_tensorrt_llm
modal run src/llm/trtllm/whisper/client.py \
    --no-streaming \
    --action asr \
    --server-url "r15.modal.host:34101"
modal run src/llm/trtllm/whisper/client.py \
    --streaming \
    --action asr \
    --server-url "r15.modal.host:34101"
modal run src/llm/trtllm/whisper/client.py \
    --action asr \
    --reference-audio /1221-135766-0001.wav \
    --text-prefix "<|startoftranscript|><|en|><|transcribe|><|notimestamps|>" \
    --server-url "r15.modal.host:34101"
modal run src/llm/trtllm/whisper/client.py \
    --action asr \
    --reference-audio /long.wav \
    --text-prefix "<|startoftranscript|><|zh|><|transcribe|><|notimestamps|>" \
    --server-url "r24.modal.host:44175"

## single wav test for whisper_infer_bls,whisper_tensorrt_llm_cpprunner
modal run src/llm/trtllm/whisper/client.py \
    --no-streaming \
    --action asr \
    --model-name whisper_infer_bls \
    --server-url "r28.modal.host:38535"
modal run src/llm/trtllm/whisper/client.py \
    --streaming \
    --action asr \
    --model-name whisper_infer_bls \
    --server-url "r15.modal.host:34101"
modal run src/llm/trtllm/whisper/client.py \
    --action asr \
    --model-name whisper_infer_bls \
    --reference-audio /1221-135766-0001.wav \
    --text-prefix "<|startoftranscript|><|en|><|transcribe|><|notimestamps|>" \
    --server-url "r15.modal.host:34101"
modal run src/llm/trtllm/whisper/client.py \
    --action asr \
    --model-name whisper_infer_bls \
    --reference-audio /long.wav \
    --text-prefix "<|startoftranscript|><|zh|><|transcribe|><|notimestamps|>" \
    --server-url "r15.modal.host:34101"

## bench for whisper
modal run src/llm/trtllm/whisper/client.py \
    --no-streaming \
    --action bench_asr \
    --model-name whisper \
    --concurency-cn 4 \
    --batch-size 4 \
    --reference-audio /long.wav \
    --text-prefix "<|startoftranscript|><|zh|><|transcribe|><|notimestamps|>" \
    --server-url "r15.modal.host:34101"

## bench (concurency_cn:1->2->4->8->16 | batch_size(requests_cn):1->2->4->8)
## bench throughput and latency, grpc just test, modal support http, grpc now use tunnel
modal run src/llm/trtllm/whisper/client.py \
    --no-streaming \
    --action bench_asr \
    --concurency-cn 4 \
    --batch-size 4 \
    --server-url "r28.modal.host:33695"

modal run src/llm/trtllm/whisper/client.py \
    --streaming \
    --action bench_asr \
    --concurency-cn 4 \
    --batch-size 4 \
    --server-url "r18.modal.host:41787"

modal run src/llm/trtllm/whisper/client.py \
    --streaming \
    --action bench_asr \
    --concurency-cn 4 \
    --batch-size 4 \
    --reference-audio /long.wav \
    --text-prefix "<|startoftranscript|><|zh|><|transcribe|><|notimestamps|>" \
    --server-url "r18.modal.host:41787"

## bench for whisper_infer_bls,whisper_tensorrt_llm_cpprunner
modal run src/llm/trtllm/whisper/client.py \
    --no-streaming \
    --action bench_asr \
    --model-name whisper_infer_bls \
    --concurency-cn 4 \
    --batch-size 4 \
    --reference-audio /long.wav \
    --text-prefix "<|startoftranscript|><|zh|><|transcribe|><|notimestamps|>" \
    --server-url "r15.modal.host:34101"

# http
## health check
TRITON_PROTOCOL=http modal run src/llm/trtllm/whisper/client.py \
    --action health --verbose \
    --server-url "weedge--tritonserver-serve-dev.modal.run"

## single wav test
TRITON_PROTOCOL=http modal run src/llm/trtllm/whisper/client.py \
    --no-streaming \
    --action asr \
    --server-url "weedge--tritonserver-serve-dev.modal.run"

## bench (concurency_cn:1->2->4->8->16 | batch_size(requests_cn):1->2->4->8)
## bench throughput and latency, grpc just test, modal support http, grpc now use tunnel
TRITON_PROTOCOL=http modal run src/llm/trtllm/whisper/client.py \
    --no-streaming \
    --action bench_asr \
    --concurency-cn 4 \
    --batch-size 4 \
    --server-url "weedge--tritonserver-serve-dev.modal.run"

# other the same as grpc :-)

# WER eval
see run.py to change
"""


@app.local_entrypoint()
def main(
    server_url: str = "localhost:8000",
    reference_audio: str = "/1221-135766-0002.wav",
    text_prefix: str = "<|startoftranscript|><|en|><|transcribe|><|notimestamps|>",
    model_name: str = "whisper_bls",
    action: str = "health",
    concurency_cn: int = 1,
    batch_size: int = 1,
    padding_duration: int = 10,  # padding to nearset 10 seconds , max 30 seconds
    verbose: bool = False,
    streaming: bool = False,
    model_names: str = "whisper_bls,whisper_tensorrt_llm",
):
    if action == "asr":
        asr.remote(
            server_url,
            reference_audio,
            text_prefix,
            model_name,
            padding_duration,
            verbose=verbose,
            streaming=streaming,
        )
    elif action == "bench_asr":
        """
        bench throughput and latency, grpc just test, modal support http, grpc now use tunnel
        """
        bench_asr.remote(
            server_url,
            reference_audio,
            text_prefix,
            model_name,
            padding_duration,
            concurency_cn,
            batch_size,
            verbose=verbose,
            streaming=streaming,
        )
    else:
        health.remote(server_url, verbose=verbose, model_names=model_names)


def store_transcripts(filename: os.PathLike, texts: Iterable[Tuple[str, str, str]]) -> None:
    """Save predicted results and reference transcripts to a file.
    https://github.com/k2-fsa/icefall/blob/master/icefall/utils.py
    Args:
      filename:
        File to save the results to.
      texts:
        An iterable of tuples. The first element is the cur_id, the second is
        the reference transcript and the third element is the predicted result.
    Returns:
      Return None.
    """
    with open(filename, "w") as f:
        for cut_id, ref, hyp in texts:
            print(f"{cut_id}:\tref={ref}", file=f)
            print(f"{cut_id}:\thyp={hyp}", file=f)


def write_error_stats(
    f: TextIO,
    test_set_name: str,
    results: List[Tuple[str, str, str]],  # wav_id, label, prediction
    enable_log: bool = True,
) -> float:
    """Write statistics based on predicted results and reference transcripts.
    https://github.com/k2-fsa/icefall/blob/master/icefall/utils.py
    It will write the following to the given file:

        - WER
        - number of insertions, deletions, substitutions, corrects and total
          reference words. For example::

              Errors: 23 insertions, 57 deletions, 212 substitutions, over 2606
              reference words (2337 correct)

        - The difference between the reference transcript and predicted result.
          An instance is given below::

            THE ASSOCIATION OF (EDISON->ADDISON) ILLUMINATING COMPANIES

          The above example shows that the reference word is `EDISON`,
          but it is predicted to `ADDISON` (a substitution error).

          Another example is::

            FOR THE FIRST DAY (SIR->*) I THINK

          The reference word `SIR` is missing in the predicted
          results (a deletion error).
      results:
        An iterable of tuples. The first element is the cur_id, the second is
        the reference transcript and the third element is the predicted result.
      enable_log:
        If True, also print detailed WER to the console.
        Otherwise, it is written only to the given file.
    Returns:
      Return None.
    """
    subs: Dict[Tuple[str, str], int] = defaultdict(int)
    ins: Dict[str, int] = defaultdict(int)
    dels: Dict[str, int] = defaultdict(int)

    # `words` stores counts per word, as follows:
    #   corr, ref_sub, hyp_sub, ins, dels
    words: Dict[str, List[int]] = defaultdict(lambda: [0, 0, 0, 0, 0])
    num_corr = 0
    ERR = "*"
    for cut_id, ref, hyp in results:
        ali = kaldialign.align(ref, hyp, ERR)
        for ref_word, hyp_word in ali:
            if ref_word == ERR:
                ins[hyp_word] += 1
                words[hyp_word][3] += 1
            elif hyp_word == ERR:
                dels[ref_word] += 1
                words[ref_word][4] += 1
            elif hyp_word != ref_word:
                subs[(ref_word, hyp_word)] += 1
                words[ref_word][1] += 1
                words[hyp_word][2] += 1
            else:
                words[ref_word][0] += 1
                num_corr += 1
    ref_len = sum([len(r) for _, r, _ in results])
    sub_errs = sum(subs.values())
    ins_errs = sum(ins.values())
    del_errs = sum(dels.values())
    tot_errs = sub_errs + ins_errs + del_errs
    tot_err_rate = "%.2f" % (100.0 * tot_errs / ref_len)

    if enable_log:
        logging.info(
            f"[{test_set_name}] %WER {tot_errs / ref_len:.2%} "
            f"[{tot_errs} / {ref_len}, {ins_errs} ins, "
            f"{del_errs} del, {sub_errs} sub ]"
        )

    print(f"%WER = {tot_err_rate}", file=f)
    print(
        f"Errors: {ins_errs} insertions, {del_errs} deletions, "
        f"{sub_errs} substitutions, over {ref_len} reference "
        f"words ({num_corr} correct)",
        file=f,
    )
    print(
        "Search below for sections starting with PER-UTT DETAILS:, "
        "SUBSTITUTIONS:, DELETIONS:, INSERTIONS:, PER-WORD STATS:",
        file=f,
    )

    print("", file=f)
    print("PER-UTT DETAILS: corr or (ref->hyp)  ", file=f)
    for cut_id, ref, hyp in results:
        ali = kaldialign.align(ref, hyp, ERR)
        combine_successive_errors = True
        if combine_successive_errors:
            ali = [[[x], [y]] for x, y in ali]
            for i in range(len(ali) - 1):
                if ali[i][0] != ali[i][1] and ali[i + 1][0] != ali[i + 1][1]:
                    ali[i + 1][0] = ali[i][0] + ali[i + 1][0]
                    ali[i + 1][1] = ali[i][1] + ali[i + 1][1]
                    ali[i] = [[], []]
            ali = [
                [
                    list(filter(lambda a: a != ERR, x)),
                    list(filter(lambda a: a != ERR, y)),
                ]
                for x, y in ali
            ]
            ali = list(filter(lambda x: x != [[], []], ali))
            ali = [
                [
                    ERR if x == [] else " ".join(x),
                    ERR if y == [] else " ".join(y),
                ]
                for x, y in ali
            ]

        print(
            f"{cut_id}:\t"
            + " ".join(
                (
                    ref_word if ref_word == hyp_word else f"({ref_word}->{hyp_word})"
                    for ref_word, hyp_word in ali
                )
            ),
            file=f,
        )

    print("", file=f)
    print("SUBSTITUTIONS: count ref -> hyp", file=f)

    for count, (ref, hyp) in sorted([(v, k) for k, v in subs.items()], reverse=True):
        print(f"{count}   {ref} -> {hyp}", file=f)

    print("", file=f)
    print("DELETIONS: count ref", file=f)
    for count, ref in sorted([(v, k) for k, v in dels.items()], reverse=True):
        print(f"{count}   {ref}", file=f)

    print("", file=f)
    print("INSERTIONS: count hyp", file=f)
    for count, hyp in sorted([(v, k) for k, v in ins.items()], reverse=True):
        print(f"{count}   {hyp}", file=f)

    print("", file=f)
    print("PER-WORD STATS: word  corr tot_errs count_in_ref count_in_hyp", file=f)
    for _, word, counts in sorted([(sum(v[1:]), k, v) for k, v in words.items()], reverse=True):
        (corr, ref_sub, hyp_sub, ins, dels) = counts
        tot_errs = ref_sub + hyp_sub + ins + dels
        ref_count = corr + ref_sub + dels
        hyp_count = corr + hyp_sub + ins

        print(f"{word}   {corr} {tot_errs} {ref_count} {hyp_count}", file=f)
    return float(tot_err_rate)



import io
import os
import logging
from typing import AsyncGenerator

import aiohttp
from PIL import Image
from apipeline.frames.data_frames import Frame, ImageRawFrame
from apipeline.frames.sys_frames import ErrorFrame

from src.processors.image.base import ImageGenProcessor


class ComfyUIAPIImageGenProcessor(ImageGenProcessor):
    """
    ComfyUI API Image Gen Processor
    need cd deploy/modal to run src/comfyui/server.py
    ```
    MODEL_NAME=flux1_schnell_fp8 IMAGE_GPU=L40S modal serve src/comfyui/server.py
    ```
    """

    def __init__(
        self,
        *,
        aiohttp_session: aiohttp.ClientSession = None,
        api_url: str = "",
        model: str = "flux1_schnell_fp8",
        width: int = 1024,
        height: int = 1024,
        steps: int = 4,  # The number of steps used in the denoising process.
    ):
        super().__init__()
        self._model = model
        self._width = width
        self._height = height
        self._steps = steps
        self._aiohttp_session = aiohttp_session
        self._api_url = api_url or os.getenv("API_URL", "")

    async def run_image_gen(self, prompt: str) -> AsyncGenerator[Frame, None]:
        logging.debug(f"Generating image from prompt: {prompt}")

        response = await self._aiohttp_session.post(
            self._api_url,
            json={
                "model": self._model,
                "prompt": prompt,
                "width": self._width,
                "height": self._height,
                "steps": self._steps,
            },
        )
        image_bytes = await response.content.read()

        if not image_bytes:
            yield ErrorFrame("Image generation failed")
            return

        image_stream = io.BytesIO(image_bytes)
        image = Image.open(image_stream).convert("RGB")
        # if image.size != (self._width, self._height):
        #    image = image.resize((self._width, self._height))
        frame = ImageRawFrame(
            image=image.tobytes(),
            size=image.size,
            format=image.format if image.format else "JPEG",
            mode=image.mode,
        )
        yield frame



# Copyright (c) 2024 Alibaba Inc (authors: Xiang Lyu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
import onnxruntime
import torch
import numpy as np
import whisper
from typing import Callable
import torchaudio.compliance.kaldi as kaldi
import torchaudio
import os
import re
import inflect
try:
    import ttsfrd
    use_ttsfrd = True
except ImportError:
    print("failed to import ttsfrd, use WeTextProcessing instead")
    from tn.chinese.normalizer import Normalizer as ZhNormalizer
    from tn.english.normalizer import Normalizer as EnNormalizer
    use_ttsfrd = False
from cosyvoice.utils.frontend_utils import contains_chinese, replace_blank, replace_corner_mark, remove_bracket, spell_out_number, split_paragraph


class CosyVoiceFrontEnd:

    def __init__(self,
                 get_tokenizer: Callable,
                 feat_extractor: Callable,
                 campplus_model: str,
                 speech_tokenizer_model: str,
                 spk2info: str = '',
                 instruct: bool = False,
                 allowed_special: str = 'all'):
        self.tokenizer = get_tokenizer()
        self.feat_extractor = feat_extractor
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        option = onnxruntime.SessionOptions()
        option.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
        option.intra_op_num_threads = 1
        self.campplus_session = onnxruntime.InferenceSession(campplus_model, sess_options=option, providers=["CPUExecutionProvider"])
        self.speech_tokenizer_session = onnxruntime.InferenceSession(speech_tokenizer_model, sess_options=option, providers=["CUDAExecutionProvider"if torch.cuda.is_available() else "CPUExecutionProvider"])
        if os.path.exists(spk2info):
            self.spk2info = torch.load(spk2info, map_location=self.device)
        self.instruct = instruct
        self.allowed_special = allowed_special
        self.inflect_parser = inflect.engine()
        self.use_ttsfrd = use_ttsfrd
        if self.use_ttsfrd:
            self.frd = ttsfrd.TtsFrontendEngine()
            ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
            assert self.frd.initialize('{}/../../pretrained_models/CosyVoice-ttsfrd/resource'.format(ROOT_DIR)) is True, 'failed to initialize ttsfrd resource'
            self.frd.set_lang_type('pinyin')
            self.frd.enable_pinyin_mix(True)
            self.frd.set_breakmodel_index(1)
        else:
            self.zh_tn_model = ZhNormalizer(remove_erhua=False, full_to_half=False)
            self.en_tn_model = EnNormalizer()

    def _extract_text_token(self, text):
        text_token = self.tokenizer.encode(text, allowed_special=self.allowed_special)
        text_token = torch.tensor([text_token], dtype=torch.int32).to(self.device)
        text_token_len = torch.tensor([text_token.shape[1]], dtype=torch.int32).to(self.device)
        return text_token, text_token_len

    def _extract_speech_token(self, speech):
        feat = whisper.log_mel_spectrogram(speech, n_mels=128)
        speech_token = self.speech_tokenizer_session.run(None, {self.speech_tokenizer_session.get_inputs()[0].name: feat.detach().cpu().numpy(),
                                                                self.speech_tokenizer_session.get_inputs()[1].name: np.array([feat.shape[2]], dtype=np.int32)})[0].flatten().tolist()
        speech_token = torch.tensor([speech_token], dtype=torch.int32).to(self.device)
        speech_token_len = torch.tensor([speech_token.shape[1]], dtype=torch.int32).to(self.device)
        return speech_token, speech_token_len

    def _extract_spk_embedding(self, speech):
        feat = kaldi.fbank(speech,
                           num_mel_bins=80,
                           dither=0,
                           sample_frequency=16000)
        feat = feat - feat.mean(dim=0, keepdim=True)
        embedding = self.campplus_session.run(None, {self.campplus_session.get_inputs()[0].name: feat.unsqueeze(dim=0).cpu().numpy()})[0].flatten().tolist()
        embedding = torch.tensor([embedding]).to(self.device)
        return embedding

    def _extract_speech_feat(self, speech):
        speech_feat = self.feat_extractor(speech).squeeze(dim=0).transpose(0, 1).to(self.device)
        speech_feat = speech_feat.unsqueeze(dim=0)
        speech_feat_len = torch.tensor([speech_feat.shape[1]], dtype=torch.int32).to(self.device)
        return speech_feat, speech_feat_len

    def text_normalize(self, text, split=True):
        text = text.strip()
        if contains_chinese(text):
            if self.use_ttsfrd:
                text = self.frd.get_frd_extra_info(text, 'input')
            else:
                text = self.zh_tn_model.normalize(text)
            text = text.replace("\n", "")
            text = replace_blank(text)
            text = replace_corner_mark(text)
            text = text.replace(".", "、")
            text = text.replace(" - ", "，")
            text = remove_bracket(text)
            text = re.sub(r'[，,]+$', '。', text)
            texts = [i for i in split_paragraph(text, partial(self.tokenizer.encode, allowed_special=self.allowed_special), "zh", token_max_n=80,
                                                token_min_n=60, merge_len=20,
                                                comma_split=False)]
        else:
            if self.use_ttsfrd:
                text = self.frd.get_frd_extra_info(text, 'input')
            else:
                text = self.en_tn_model.normalize(text)
            text = spell_out_number(text, self.inflect_parser)
            texts = [i for i in split_paragraph(text, partial(self.tokenizer.encode, allowed_special=self.allowed_special), "en", token_max_n=80,
                                                token_min_n=60, merge_len=20,
                                                comma_split=False)]
        if split is False:
            return text
        return texts

    def frontend_sft(self, tts_text, spk_id):
        tts_text_token, tts_text_token_len = self._extract_text_token(tts_text)
        embedding = self.spk2info[spk_id]['embedding']
        model_input = {'text': tts_text_token, 'text_len': tts_text_token_len, 'llm_embedding': embedding, 'flow_embedding': embedding}
        return model_input

    def frontend_zero_shot(self, tts_text, prompt_text, prompt_speech_16k):
        tts_text_token, tts_text_token_len = self._extract_text_token(tts_text)
        prompt_text_token, prompt_text_token_len = self._extract_text_token(prompt_text)
        prompt_speech_22050 = torchaudio.transforms.Resample(orig_freq=16000, new_freq=22050)(prompt_speech_16k)
        speech_feat, speech_feat_len = self._extract_speech_feat(prompt_speech_22050)
        speech_token, speech_token_len = self._extract_speech_token(prompt_speech_16k)
        embedding = self._extract_spk_embedding(prompt_speech_16k)
        model_input = {'text': tts_text_token, 'text_len': tts_text_token_len,
                       'prompt_text': prompt_text_token, 'prompt_text_len': prompt_text_token_len,
                       'llm_prompt_speech_token': speech_token, 'llm_prompt_speech_token_len': speech_token_len,
                       'flow_prompt_speech_token': speech_token, 'flow_prompt_speech_token_len': speech_token_len,
                       'prompt_speech_feat': speech_feat, 'prompt_speech_feat_len': speech_feat_len,
                       'llm_embedding': embedding, 'flow_embedding': embedding}
        return model_input

    def frontend_cross_lingual(self, tts_text, prompt_speech_16k):
        model_input = self.frontend_zero_shot(tts_text, '', prompt_speech_16k)
        # in cross lingual mode, we remove prompt in llm
        del model_input['prompt_text']
        del model_input['prompt_text_len']
        del model_input['llm_prompt_speech_token']
        del model_input['llm_prompt_speech_token_len']
        return model_input

    def frontend_instruct(self, tts_text, spk_id, instruct_text):
        model_input = self.frontend_sft(tts_text, spk_id)
        # in instruct mode, we remove spk_embedding in llm due to information leakage
        del model_input['llm_embedding']
        instruct_text_token, instruct_text_token_len = self._extract_text_token(instruct_text + '<endofprompt>')
        model_input['prompt_text'] = instruct_text_token
        model_input['prompt_text_len'] = instruct_text_token_len
        return model_input



'''
    This part of code contains a pretrained vgg_face model.
    ref link: https://github.com/prlz77/vgg-face.pytorch
'''
import torch
import torch.nn.functional as F
import torch.utils.model_zoo
import pickle
from musetalk.loss import resnet as ResNet


MODEL_URL = "https://github.com/claudio-unipv/vggface-pytorch/releases/download/v0.1/vggface-9d491dd7c30312.pth"
VGG_FACE_PATH = '/apdcephfs_cq8/share_1367250/zhentaoyu/Driving/00_VASA/00_data/models/pretrain_models/resnet50_ft_weight.pkl'

# It was 93.5940, 104.7624, 129.1863 before dividing by 255
MEAN_RGB = [
    0.367035294117647,
    0.41083294117647057,
    0.5066129411764705
]
def load_state_dict(model, fname):
    """
    Set parameters converted from Caffe models authors of VGGFace2 provide.
    See https://www.robots.ox.ac.uk/~vgg/data/vgg_face2/.

    Arguments:
        model: model
        fname: file name of parameters converted from a Caffe model, assuming the file format is Pickle.
    """
    with open(fname, 'rb') as f:
        weights = pickle.load(f, encoding='latin1')

    own_state = model.state_dict()
    for name, param in weights.items():
        if name in own_state:
            try:
                own_state[name].copy_(torch.from_numpy(param))
            except Exception:
                raise RuntimeError('While copying the parameter named {}, whose dimensions in the model are {} and whose '\
                                   'dimensions in the checkpoint are {}.'.format(name, own_state[name].size(), param.size()))
        else:
            raise KeyError('unexpected key "{}" in state_dict'.format(name))
        

def vggface2(pretrained=True):
    vggface = ResNet.resnet50(num_classes=8631, include_top=True)
    load_state_dict(vggface, VGG_FACE_PATH)
    return vggface

def vggface(pretrained=False, **kwargs):
    """VGGFace model.

    Args:
        pretrained (bool): If True, returns pre-trained model
    """
    model = VggFace(**kwargs)
    if pretrained:
        state = torch.utils.model_zoo.load_url(MODEL_URL)
        model.load_state_dict(state)
    return model


class VggFace(torch.nn.Module):
    def __init__(self, classes=2622):
        """VGGFace model.

        Face recognition network.  It takes as input a Bx3x224x224
        batch of face images and gives as output a BxC score vector
        (C is the number of identities).
        Input images need to be scaled in the 0-1 range and then
        normalized with respect to the mean RGB used during training.

        Args:
            classes (int): number of identities recognized by the
            network

        """
        super().__init__()
        self.conv1 = _ConvBlock(3, 64, 64)
        self.conv2 = _ConvBlock(64, 128, 128)
        self.conv3 = _ConvBlock(128, 256, 256, 256)
        self.conv4 = _ConvBlock(256, 512, 512, 512)
        self.conv5 = _ConvBlock(512, 512, 512, 512)
        self.dropout = torch.nn.Dropout(0.5)
        self.fc1 = torch.nn.Linear(7 * 7 * 512, 4096)
        self.fc2 = torch.nn.Linear(4096, 4096)
        self.fc3 = torch.nn.Linear(4096, classes)

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = self.conv5(x)
        x = x.view(x.size(0), -1)
        x = self.dropout(F.relu(self.fc1(x)))
        x = self.dropout(F.relu(self.fc2(x)))
        x = self.fc3(x)
        return x


class _ConvBlock(torch.nn.Module):
    """A Convolutional block."""

    def __init__(self, *units):
        """Create a block with len(units) - 1 convolutions.

        convolution number i transforms the number of channels from
        units[i - 1] to units[i] channels.

        """
        super().__init__()
        self.convs = torch.nn.ModuleList([
            torch.nn.Conv2d(in_, out, 3, 1, 1)
            for in_, out in zip(units[:-1], units[1:])
        ])

    def forward(self, x):
        # Each convolution is followed by a ReLU, then the block is
        # concluded by a max pooling.
        for c in self.convs:
            x = F.relu(c(x))
        return F.max_pool2d(x, 2, 2, 0, ceil_mode=True)



import numpy as np
from torchvision import models
class Vgg19(torch.nn.Module):
    """
    Vgg19 network for perceptual loss.
    """
    def __init__(self, requires_grad=False):
        super(Vgg19, self).__init__()
        vgg_pretrained_features = models.vgg19(pretrained=True).features
        self.slice1 = torch.nn.Sequential()
        self.slice2 = torch.nn.Sequential()
        self.slice3 = torch.nn.Sequential()
        self.slice4 = torch.nn.Sequential()
        self.slice5 = torch.nn.Sequential()
        for x in range(2):
            self.slice1.add_module(str(x), vgg_pretrained_features[x])
        for x in range(2, 7):
            self.slice2.add_module(str(x), vgg_pretrained_features[x])
        for x in range(7, 12):
            self.slice3.add_module(str(x), vgg_pretrained_features[x])
        for x in range(12, 21):
            self.slice4.add_module(str(x), vgg_pretrained_features[x])
        for x in range(21, 30):
            self.slice5.add_module(str(x), vgg_pretrained_features[x])

        self.mean = torch.nn.Parameter(data=torch.Tensor(np.array([0.485, 0.456, 0.406]).reshape((1, 3, 1, 1))),
                                       requires_grad=False)
        self.std = torch.nn.Parameter(data=torch.Tensor(np.array([0.229, 0.224, 0.225]).reshape((1, 3, 1, 1))),
                                      requires_grad=False)

        if not requires_grad:
            for param in self.parameters():
                param.requires_grad = False

    def forward(self, X):
        X = (X - self.mean) / self.std
        h_relu1 = self.slice1(X)
        h_relu2 = self.slice2(h_relu1)
        h_relu3 = self.slice3(h_relu2)
        h_relu4 = self.slice4(h_relu3)
        h_relu5 = self.slice5(h_relu4)
        out = [h_relu1, h_relu2, h_relu3, h_relu4, h_relu5]
        return out


from torch import nn
class AntiAliasInterpolation2d(nn.Module):
    """
    Band-limited downsampling, for better preservation of the input signal.
    """
    def __init__(self, channels, scale):
        super(AntiAliasInterpolation2d, self).__init__()
        sigma = (1 / scale - 1) / 2
        kernel_size = 2 * round(sigma * 4) + 1
        self.ka = kernel_size // 2
        self.kb = self.ka - 1 if kernel_size % 2 == 0 else self.ka

        kernel_size = [kernel_size, kernel_size]
        sigma = [sigma, sigma]
        # The gaussian kernel is the product of the
        # gaussian function of each dimension.
        kernel = 1
        meshgrids = torch.meshgrid(
            [
                torch.arange(size, dtype=torch.float32)
                for size in kernel_size
                ]
        )
        for size, std, mgrid in zip(kernel_size, sigma, meshgrids):
            mean = (size - 1) / 2
            kernel *= torch.exp(-(mgrid - mean) ** 2 / (2 * std ** 2))

        # Make sure sum of values in gaussian kernel equals 1.
        kernel = kernel / torch.sum(kernel)
        # Reshape to depthwise convolutional weight
        kernel = kernel.view(1, 1, *kernel.size())
        kernel = kernel.repeat(channels, *[1] * (kernel.dim() - 1))

        self.register_buffer('weight', kernel)
        self.groups = channels
        self.scale = scale
        inv_scale = 1 / scale
        self.int_inv_scale = int(inv_scale)

    def forward(self, input):
        if self.scale == 1.0:
            return input

        out = F.pad(input, (self.ka, self.kb, self.ka, self.kb))
        out = F.conv2d(out, weight=self.weight, groups=self.groups)
        out = out[:, :, ::self.int_inv_scale, ::self.int_inv_scale]

        return out


class ImagePyramide(torch.nn.Module):
    """
    Create image pyramide for computing pyramide perceptual loss.
    """
    def __init__(self, scales, num_channels):
        super(ImagePyramide, self).__init__()
        downs = {}
        for scale in scales:
            downs[str(scale).replace('.', '-')] = AntiAliasInterpolation2d(num_channels, scale)
        self.downs = nn.ModuleDict(downs)

    def forward(self, x):
        out_dict = {}
        for scale, down_module in self.downs.items():
            out_dict['prediction_' + str(scale).replace('-', '.')] = down_module(x)
        return out_dict


# Copyright (c) 2024 Alibaba Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


def build_env(config, config_name, path):
    t_path = os.path.join(path, config_name)
    if config != t_path:
        os.makedirs(path, exist_ok=True)
        shutil.copyfile(config, os.path.join(path, config_name))
